"""Test some of the core functionality of the Housekeeper API."""

import pytest
from sqlalchemy.exc import OperationalError

from cg.apps.housekeeper.hk import HousekeeperAPI


def test_non_initialised_db(hk_config: dict, hk_tag: str):
    """Test to use a database that is not initialised."""
    # GIVEN a housekeeper api and some Housekeeper configs
    api = HousekeeperAPI(hk_config)

    # GIVEN a api without the database
    with pytest.raises(OperationalError):
        # THEN it should raise a operational error
        api.add_tag(hk_tag)


def test_init_db(hk_config: dict, hk_tag: str):
    """Test to setup the database."""
    # GIVEN a Housekeeper API and a Housekeeper config
    api = HousekeeperAPI(hk_config)

    # WHEN initiating the database
    api.initialise_db()

    # THEN the api should not throw an exception
    assert api.add_tag(hk_tag)



from http import HTTPStatus

import mock.mock
import pytest
from flask.testing import FlaskClient

from cg.apps.tb import TrailblazerAPI
from cg.apps.tb.dto.summary_response import AnalysisSummary
from cg.constants.constants import Workflow
from cg.store.models import Order


@pytest.mark.parametrize(
    "limit, workflow, expected_orders",
    [
        (None, Workflow.MIP_DNA, 2),
        (1, Workflow.MIP_DNA, 1),
        (2, Workflow.MIP_DNA, 2),
        (None, Workflow.BALSAMIC, 1),
        (None, Workflow.FLUFFY, 0),
        (None, None, 3),
    ],
)
def test_orders_endpoint(
    client: FlaskClient,
    order: Order,
    order_another: Order,
    order_balsamic: Order,
    limit: int | None,
    workflow: str,
    expected_orders: int,
):
    """Tests that orders are returned from the orders endpoint"""
    # GIVEN a store with three orders, two of which are MIP-DNA and the last is BALSAMIC

    # WHEN a request is made to get all orders
    endpoint: str = "/api/v1/orders"
    with mock.patch.object(
        TrailblazerAPI,
        "get_summaries",
        return_value=[
            AnalysisSummary(
                order_id=order.id, cancelled=0, completed=1, delivered=0, failed=0, running=0
            ),
            AnalysisSummary(
                order_id=order_another.id,
                cancelled=0,
                completed=1,
                delivered=0,
                failed=0,
                running=0,
            ),
            AnalysisSummary(
                order_id=order_balsamic.id,
                cancelled=0,
                completed=1,
                delivered=0,
                failed=0,
                running=0,
            ),
        ],
    ):
        response = client.get(endpoint, query_string={"pageSize": limit, "workflow": workflow})

    # THEN the response should be successful
    assert response.status_code == HTTPStatus.OK

    # THEN the response contains the correct number of orders
    assert len(response.json["orders"]) == expected_orders


def test_order_endpoint(
    client: FlaskClient, order: Order, order_another: Order, analysis_summary: AnalysisSummary
):
    """Tests that the order endpoint returns the order with matching id"""
    # GIVEN a store with two orders
    order_id_to_fetch: int = order.id

    # WHEN a request is made to get a specific order
    endpoint: str = f"/api/v1/orders/{order_id_to_fetch}"
    with mock.patch.object(TrailblazerAPI, "get_summaries", return_value=[analysis_summary]):
        response = client.get(endpoint)

    # THEN the response should be successful
    assert response.status_code == HTTPStatus.OK

    # THEN the response contains the specified order
    assert response.json["id"] == order_id_to_fetch


def test_order_endpoint_not_found(
    client: FlaskClient, order: Order, order_another: Order, non_existent_order_id: int
):
    """Tests that the order endpoint returns the order with matching id"""
    # GIVEN a store with two orders

    # WHEN a request is made to get a non-existent order
    endpoint: str = f"/api/v1/orders/{non_existent_order_id}"
    response = client.get(endpoint)

    # THEN the response should be unsuccessful
    assert response.status_code == HTTPStatus.NOT_FOUND



"""Module for modeling flow cells."""

import datetime
import logging
import os
from pathlib import Path
from typing import Type

from typing_extensions import Literal

from cg.apps.demultiplex.sample_sheet.sample_models import (
    FlowCellSampleBcl2Fastq,
    FlowCellSampleBCLConvert,
)
from cg.apps.demultiplex.sample_sheet.sample_sheet_models import SampleSheet
from cg.apps.demultiplex.sample_sheet.sample_sheet_validator import SampleSheetValidator
from cg.cli.demultiplex.copy_novaseqx_demultiplex_data import get_latest_analysis_path
from cg.constants.constants import LENGTH_LONG_DATE
from cg.constants.demultiplexing import BclConverter, DemultiplexingDirsAndFiles
from cg.constants.sequencing import SEQUENCER_TYPES, Sequencers
from cg.constants.symbols import EMPTY_STRING
from cg.exc import FlowCellError
from cg.models.demultiplex.run_parameters import (
    RunParameters,
    RunParametersHiSeq,
    RunParametersNovaSeq6000,
    RunParametersNovaSeqX,
)
from cg.models.flow_cell.utils import parse_date

LOG = logging.getLogger(__name__)
RUN_PARAMETERS_CONSTRUCTOR: dict[str, Type] = {
    Sequencers.HISEQGA: RunParametersHiSeq,
    Sequencers.HISEQX: RunParametersHiSeq,
    Sequencers.NOVASEQ: RunParametersNovaSeq6000,
    Sequencers.NOVASEQX: RunParametersNovaSeqX,
}


class FlowCellDirectoryData:
    """Class to collect information about flow cell directories and their particular files."""

    def __init__(self, flow_cell_path: Path, bcl_converter: str | None = None):
        LOG.debug(f"Instantiating FlowCellDirectoryData with path {flow_cell_path}")
        self.path: Path = flow_cell_path
        self.machine_name: str = EMPTY_STRING
        self._run_parameters: RunParameters | None = None
        self.run_date: datetime.datetime = datetime.datetime.now()
        self.machine_number: int = 0
        self.base_name: str = EMPTY_STRING  # Base name is flow cell-id + flow cell position
        self.id: str = EMPTY_STRING
        self.position: Literal["A", "B"] = "A"
        self.parse_flow_cell_dir_name()
        self.bcl_converter: str = bcl_converter or BclConverter.BCLCONVERT
        self._sample_sheet_path_hk: Path | None = None
        self.sample_sheet_validator = SampleSheetValidator()

    def parse_flow_cell_dir_name(self):
        """Parse relevant information from flow cell name.
        This will assume that the flow cell naming convention is used. If not we skip the flow cell.
        Convention is: <date>_<machine>_<run_numbers>_<A|B><flow_cell_id>
        Example: '201203_D00483_0200_AHVKJCDRXX'.
        """

        self.validate_flow_cell_dir_name()
        self.run_date = self._parse_date()
        self.machine_name = self.split_flow_cell_name[1]
        self.machine_number = int(self.split_flow_cell_name[2])
        base_name: str = self.split_flow_cell_name[-1]
        self.base_name = base_name
        LOG.debug(f"Set flow cell id to {base_name}")
        self.id = base_name[1:]
        self.position = base_name[0]

    @property
    def split_flow_cell_name(self) -> list[str]:
        """Return split flow cell name."""
        return self.path.name.split("_")

    @property
    def full_name(self) -> str:
        """Return flow cell full name."""
        return self.path.name

    @property
    def sequenced_at(self) -> list[str]:
        """Return the sequencing date for the flow cell."""
        date_part: str = self.full_name.split("_")[0]
        return parse_date(date_part)

    @property
    def sample_sheet_path(self) -> Path:
        """
        Return sample sheet path.
        """
        return Path(self.path, DemultiplexingDirsAndFiles.SAMPLE_SHEET_FILE_NAME)

    def set_sample_sheet_path_hk(self, hk_path: Path):
        self._sample_sheet_path_hk = hk_path

    def get_sample_sheet_path_hk(self) -> Path:
        if not self._sample_sheet_path_hk:
            raise FlowCellError("Attribute _sample_sheet_path_hk has not been assigned yet")
        return self._sample_sheet_path_hk

    @property
    def run_parameters_path(self) -> Path:
        """Return path to run parameters file if it exists.
        Raises:
            FlowCellError if the flow cell has no run parameters file."""
        if DemultiplexingDirsAndFiles.RUN_PARAMETERS_PASCAL_CASE in os.listdir(self.path):
            return Path(self.path, DemultiplexingDirsAndFiles.RUN_PARAMETERS_PASCAL_CASE)
        elif DemultiplexingDirsAndFiles.RUN_PARAMETERS_CAMEL_CASE in os.listdir(self.path):
            return Path(self.path, DemultiplexingDirsAndFiles.RUN_PARAMETERS_CAMEL_CASE)
        else:
            message: str = f"No run parameters file found in flow cell {self.path}"
            LOG.error(message)
            raise FlowCellError(message)

    @property
    def run_parameters(self) -> RunParameters:
        """Return run parameters object."""
        if not self._run_parameters:
            self._run_parameters = RUN_PARAMETERS_CONSTRUCTOR[self.sequencer_type](
                run_parameters_path=self.run_parameters_path
            )
        return self._run_parameters

    @property
    def sample_type(
        self,
    ) -> Type[FlowCellSampleBcl2Fastq] | Type[FlowCellSampleBCLConvert]:
        """Return the sample class used in the flow cell."""
        if self.bcl_converter == BclConverter.BCL2FASTQ:
            return FlowCellSampleBcl2Fastq
        return FlowCellSampleBCLConvert

    @property
    def sequencer_type(
        self,
    ) -> Literal[Sequencers.HISEQX, Sequencers.HISEQGA, Sequencers.NOVASEQ, Sequencers.NOVASEQX]:
        """Return the sequencer type."""
        return SEQUENCER_TYPES[self.machine_name]

    @property
    def rta_complete_path(self) -> Path:
        """Return RTAComplete path."""
        return Path(self.path, DemultiplexingDirsAndFiles.RTACOMPLETE)

    @property
    def copy_complete_path(self) -> Path:
        """Return copy complete path."""
        return Path(self.path, DemultiplexingDirsAndFiles.COPY_COMPLETE)

    @property
    def hiseq_x_copy_complete_path(self) -> Path:
        """Return copy complete path for HiSeqX."""
        return Path(self.path, DemultiplexingDirsAndFiles.HISEQ_X_COPY_COMPLETE)

    @property
    def hiseq_x_delivery_started_path(self) -> Path:
        """Return delivery started path for HiSeqX."""
        return Path(self.path, DemultiplexingDirsAndFiles.DELIVERY)

    @property
    def demultiplexing_started_path(self) -> Path:
        """Return demux started path."""
        return Path(self.path, DemultiplexingDirsAndFiles.DEMUX_STARTED)

    @property
    def trailblazer_config_path(self) -> Path:
        """Return file to SLURM job ids path."""
        return Path(self.path, "slurm_job_ids.yaml")

    @property
    def is_demultiplexing_complete(self) -> bool:
        return Path(self.path, DemultiplexingDirsAndFiles.DEMUX_COMPLETE).exists()

    def _parse_date(self):
        """Return the parsed date in the correct format."""
        if len(self.split_flow_cell_name[0]) == LENGTH_LONG_DATE:
            return datetime.datetime.strptime(self.split_flow_cell_name[0], "%Y%m%d")
        return datetime.datetime.strptime(self.split_flow_cell_name[0], "%y%m%d")

    def validate_flow_cell_dir_name(self) -> None:
        """
        Validate on the following criteria:
        Convention is: <date>_<machine>_<run_numbers>_<A|B><flow_cell_id>
        Example: '201203_D00483_0200_AHVKJCDRXX'.
        """
        if len(self.split_flow_cell_name) != 4:
            message = f"Flowcell {self.full_name} does not follow the flow cell naming convention"
            LOG.warning(message)
            raise FlowCellError(message)

    def has_demultiplexing_started_locally(self) -> bool:
        """Check if demultiplexing has started path exists on the cluster."""
        return self.demultiplexing_started_path.exists()

    def has_demultiplexing_started_on_sequencer(self) -> bool:
        """Check if demultiplexing has started on the NovaSeqX machine."""
        latest_analysis: Path = get_latest_analysis_path(self.path)
        if not latest_analysis:
            return False
        return Path(
            latest_analysis, DemultiplexingDirsAndFiles.DATA, DemultiplexingDirsAndFiles.BCL_CONVERT
        ).exists()

    def sample_sheet_exists(self) -> bool:
        """Check if sample sheet exists."""
        LOG.info("Check if sample sheet exists")
        return self.sample_sheet_path.exists()

    @property
    def sample_sheet(self) -> SampleSheet:
        """Return sample sheet object."""
        if not self._sample_sheet_path_hk:
            raise FlowCellError("Sample sheet path has not been assigned yet")
        return self.sample_sheet_validator.get_sample_sheet_object_from_file(
            file_path=self._sample_sheet_path_hk
        )

    def is_sequencing_done(self) -> bool:
        """Check if sequencing is done.
        This is indicated by that the file RTAComplete.txt exists.
        """
        LOG.info("Check if sequencing is done")
        return self.rta_complete_path.exists()

    def is_copy_completed(self) -> bool:
        """Check if copy of flow cell is done.
        This is indicated by that the file CopyComplete.txt exists.
        """
        LOG.info("Check if copy of data from sequence instrument is ready")
        return self.copy_complete_path.exists()

    def is_flow_cell_ready(self) -> bool:
        """Check if a flow cell is ready for downstream processing.

        A flow cell is ready if the two files RTAComplete.txt and CopyComplete.txt exist in the
        flow cell directory.
        """
        LOG.info("Check if flow cell is ready for downstream processing")
        if not self.is_sequencing_done():
            LOG.warning(f"Sequencing is not completed for flow cell {self.id}")
            return False
        LOG.debug(f"Sequence is done for flow cell {self.id}")
        if not self.is_copy_completed():
            LOG.warning(f"Copy of sequence data is not ready for flow cell {self.id}")
            return False
        LOG.debug(f"All data has been transferred for flow cell {self.id}")
        LOG.info(f"Flow cell {self.id} is ready for downstream processing")
        return True

    def __str__(self):
        return f"FlowCell(path={self.path},run_parameters_path={self.run_parameters_path})"


def get_flow_cells_from_path(flow_cells_dir: Path) -> list[FlowCellDirectoryData]:
    """Return flow cell objects from flow cell dir."""
    flow_cells: list[FlowCellDirectoryData] = []
    LOG.debug(f"Search for flow cells ready to encrypt in {flow_cells_dir}")
    for flow_cell_dir in flow_cells_dir.iterdir():
        if not flow_cell_dir.is_dir():
            continue
        LOG.debug(f"Found directory: {flow_cell_dir}")
        try:
            flow_cell = FlowCellDirectoryData(flow_cell_path=flow_cell_dir)
        except FlowCellError:
            continue
        flow_cells.append(flow_cell)
    return flow_cells



from pathlib import Path

import pytest
from click.testing import CliRunner, Result

from cg.cli.demultiplex.sample_sheet import validate_sample_sheet
from cg.constants import EXIT_SUCCESS, FileExtensions
from cg.models.cg_config import CGConfig


def test_validate_non_existing_sample_sheet(
    cli_runner: CliRunner,
    sample_sheet_context: CGConfig,
):
    """Test validate sample sheet when sample sheet does not exist."""

    # GIVEN a cli runner
    # GIVEN a sample sheet file that does not exist
    sample_sheet: Path = Path("a_sample_sheet_that_does_not_exist.csv")
    assert sample_sheet.exists() is False

    # WHEN validating the sample sheet
    result = cli_runner.invoke(
        validate_sample_sheet,
        [str(sample_sheet)],
        obj=sample_sheet_context,
    )

    # THEN assert that it exits with a non-zero exit code
    assert result.exit_code != EXIT_SUCCESS
    # THEN assert the correct information was communicated
    assert f"File '{sample_sheet.name}' does not exist" in result.output


def test_validate_sample_sheet_wrong_file_type(
    cli_runner: CliRunner,
    novaseq_6000_run_parameters_path: Path,
    sample_sheet_context: CGConfig,
    caplog,
):
    """Test validate sample sheet when sample sheet is in wrong format."""
    # GIVEN an existing file in the wrong file format
    sample_sheet: Path = novaseq_6000_run_parameters_path
    assert sample_sheet.exists()
    assert sample_sheet.suffix != FileExtensions.CSV

    # WHEN validating the sample sheet
    result: Result = cli_runner.invoke(
        validate_sample_sheet,
        [str(sample_sheet)],
        obj=sample_sheet_context,
    )

    # THEN assert it exits with a non-zero exit code
    assert result.exit_code != EXIT_SUCCESS

    # THEN assert the correct information was communicated
    assert "seems to be in wrong format" in caplog.text


@pytest.mark.parametrize(
    "sample_sheet_path",
    [
        "hiseq_x_single_index_sample_sheet_path",
        "hiseq_x_dual_index_sample_sheet_path",
        "hiseq_2500_dual_index_sample_sheet_path",
        "hiseq_2500_custom_index_sample_sheet_path",
        "novaseq_6000_pre_1_5_kits_correct_sample_sheet_path",
        "novaseq_6000_post_1_5_kits_correct_sample_sheet_path",
        "novaseq_x_correct_sample_sheet",
    ],
)
def test_validate_correct_bcl_convert_sample_sheet(
    cli_runner: CliRunner,
    sample_sheet_path: str,
    sample_sheet_context: CGConfig,
    request: pytest.FixtureRequest,
):
    """Test validate sample sheet when using a BCLconvert sample sheet."""

    # GIVEN the path to a correct BCLConvert sample sheet that exists
    sample_sheet: Path = request.getfixturevalue(sample_sheet_path)
    assert sample_sheet.exists()

    # WHEN validating the sample sheet
    result: Result = cli_runner.invoke(
        validate_sample_sheet, [str(sample_sheet)], obj=sample_sheet_context
    )

    # THEN assert that it exits successfully
    assert result.exit_code == EXIT_SUCCESS



from pathlib import Path

from click.testing import CliRunner
from mock import mock

from cg.cli.workflow.mip.base import managed_variants
from cg.constants.scout import ScoutExportFileName
from cg.io.txt import read_txt
from cg.meta.workflow.mip import MipAnalysisAPI
from cg.models.cg_config import CGConfig
from tests.conftest import create_process_response


def test_managed_variants_is_written(
    case_id: str,
    cli_runner: CliRunner,
    mip_dna_context: CGConfig,
    scout_export_manged_variants_output: str,
):
    # GIVEN an analysis API
    analysis_api: MipAnalysisAPI = mip_dna_context.meta_apis["analysis_api"]

    # GIVEN a case

    # GIVEN that, the Scout command writes the managed variants to stdout
    with mock.patch(
        "cg.utils.commands.subprocess.run",
        return_value=create_process_response(std_out=scout_export_manged_variants_output),
    ):
        # WHEN creating a managed_variants file
        cli_runner.invoke(managed_variants, [case_id], obj=mip_dna_context)

    managed_variants_file = Path(analysis_api.root, case_id, ScoutExportFileName.MANAGED_VARIANTS)

    # THEN the file should exist
    assert managed_variants_file.exists()

    # THEN the file should contain the output from Scout
    file_content: str = read_txt(file_path=managed_variants_file, read_to_string=True)
    assert file_content == scout_export_manged_variants_output


def test_managed_variants_dry_run(
    case_id: str,
    cli_runner: CliRunner,
    mip_dna_context: CGConfig,
    scout_export_manged_variants_output: str,
):
    # GIVEN a case

    # GIVEN that, the Scout command writes the managed variants to stdout
    with mock.patch(
        "cg.utils.commands.subprocess.run",
        return_value=create_process_response(std_out=scout_export_manged_variants_output),
    ):
        # WHEN creating a managed_variants file using dry run
        result = cli_runner.invoke(managed_variants, [case_id, "--dry-run"], obj=mip_dna_context)

    # THEN the result should contain the output from Scout
    assert result.stdout.strip() == scout_export_manged_variants_output



from http import HTTPStatus
from unittest import mock

from flask.testing import FlaskClient

from cg.apps.tb import TrailblazerAPI
from cg.apps.tb.models import TrailblazerAnalysis
from cg.store.models import Case, Order


def test_get_delivery_message_single_case(client: FlaskClient, server_case: Case, order: Order):
    # GIVEN a case

    # WHEN a request is made to get a delivery message for the case
    response = client.get(f"/api/v1/cases/{server_case.internal_id}/delivery_message")

    # THEN the response should be successful
    assert response.status_code == HTTPStatus.OK

    # THEN the response contains a message
    assert response.json["message"]


def test_get_delivery_message_matching_order(
    client: FlaskClient, server_case: Case, server_case_in_same_order: Case, order: Order
):
    # GIVEN a case

    # WHEN a request is made to get a delivery message for the case
    response = client.get(
        f"/api/v1/cases/delivery_message?case_ids={server_case.internal_id},{server_case_in_same_order.internal_id}"
    )

    # THEN the response should be successful
    assert response.status_code == HTTPStatus.OK

    # THEN the response contains a message
    assert response.json["message"]


def test_get_delivery_message_order(
    client: FlaskClient,
    server_case: Case,
    trailblazer_analysis_for_server_case: TrailblazerAnalysis,
    server_case_in_same_order: Case,
    trailblazer_analysis_for_server_case_in_same_order: TrailblazerAnalysis,
    order: Order,
):
    """Tests that a delivery message is generated for an order with cases ready for delivery."""
    # GIVEN an order

    # WHEN a request is made to get a delivery message for the order

    # WHEN cases are ready to be delivered
    with mock.patch.object(
        TrailblazerAPI,
        "query_trailblazer",
        return_value={
            "analyses": [
                trailblazer_analysis_for_server_case.model_dump(),
                trailblazer_analysis_for_server_case_in_same_order.model_dump(),
            ]
        },
    ):
        response = client.get(f"/api/v1/orders/{order.id}/delivery_message")

        # THEN the response should be successful
        assert response.status_code == HTTPStatus.OK
        assert response.json["message"]


def test_get_delivery_message_order_missing_cases(
    client: FlaskClient,
    server_case: Case,
    trailblazer_analysis_for_server_case: TrailblazerAnalysis,
    server_case_in_same_order: Case,
    trailblazer_analysis_for_server_case_in_same_order: TrailblazerAnalysis,
    order: Order,
):
    """Tests that an error is returned for orders without any cases ready for delivery."""
    # GIVEN an order

    # WHEN a request is made to get a delivery message for the order

    # WHEN no cases are ready to be delivered
    with mock.patch.object(TrailblazerAPI, "query_trailblazer", return_value={"analyses": []}):
        response = client.get(f"/api/v1/orders/{order.id}/delivery_message")

        # THEN the response should not be successful
        assert response.status_code == HTTPStatus.PRECONDITION_FAILED
        assert response.json["error"]



"""Class to hold information about file objects"""

import gzip
import logging
from pathlib import Path
from typing import Iterable

GZIP_MAGIC_NUMBER = "1f8b"

LOG = logging.getLogger(__name__)


class FileData:
    """Helper functions for working with files"""

    @staticmethod
    def is_gzipped(path: Path) -> bool:
        """Check if file is zipped"""
        LOG.info("Check if file %s is gzipped", path)
        handle = open(path, "rb")
        if handle.read(2).hex() == GZIP_MAGIC_NUMBER:
            LOG.info("File %s is gzipped!", path)
            return True
        LOG.info("File %s is not gzipped!", path)
        return False

    @staticmethod
    def is_empty(path: Path) -> bool:
        """Check if file is empty"""
        LOG.info("Check if file %s is empty", path)
        if FileData.is_gzipped(path):
            return FileData.is_gzipped_empty(path)

        size = path.stat().st_size
        if size == 0:
            LOG.info("File %s is empty!", path)
            return True
        LOG.info("File %s is not empty", path)
        return False

    @staticmethod
    def get_file_content(path: Path) -> str:
        """Read the content of a file and return it"""
        handle = FileData.get_file_handle(path)
        return handle.read()

    @staticmethod
    def is_gzipped_empty(path: Path) -> bool:
        """Check if gzipped file is empty

        To make sure that the file is really empty we read small files to check if there are
        any content.
        If the file has more content that just metadata it is not empty.
        """
        size = path.stat().st_size
        if size > 100:
            LOG.info("File %s is not empty", path)
            return False

        content = FileData.get_file_content(path)

        if content:
            LOG.info("File %s is not empty", path)
            return False
        LOG.info("File %s is empty", path)
        return True

    @staticmethod
    def get_file_handle(path: Path) -> Iterable[str]:
        """Return the file handle of a path"""
        if FileData.is_gzipped(path):
            return gzip.open(path, "r")
        return open(path, "r")



from datetime import datetime

import pytest
from mock import Mock

from cg.apps.tb.dto.summary_response import AnalysisSummary
from cg.services.orders.order_status_service.order_summary_service import (
    OrderSummaryService,
)
from cg.store.models import Case, Customer, Order, Sample
from cg.store.store import Store
from tests.store_helpers import StoreHelpers


@pytest.fixture
def summary_service(store: Store, order: Order):
    service = OrderSummaryService(analysis_client=Mock(), store=store)
    service.analysis_client.get_summaries.return_value = [
        AnalysisSummary(
            order_id=order.id, cancelled=0, completed=1, delivered=0, failed=0, running=0
        )
    ]
    return service


@pytest.fixture
def order(store: Store, helpers: StoreHelpers) -> Order:
    customer: Customer = helpers.ensure_customer(store)
    order: Order = helpers.add_order(store=store, customer_id=customer.id, ticket_id="ticket_id")
    return order


@pytest.fixture
def sample_not_received(store: Store, helpers: StoreHelpers) -> Sample:
    return helpers.add_sample(
        store=store,
        internal_id="not_received",
    )


@pytest.fixture
def another_sample_not_received(store: Store, helpers: StoreHelpers) -> Sample:
    return helpers.add_sample(
        store=store,
        internal_id="another_not_received",
    )


@pytest.fixture
def sample_in_preparation(store: Store, helpers: StoreHelpers) -> Sample:
    return helpers.add_sample(
        store=store,
        internal_id="in_prep",
        received_at=datetime.now(),
    )


@pytest.fixture
def sample_in_sequencing(store: Store, helpers: StoreHelpers) -> Sample:
    return helpers.add_sample(
        store=store,
        internal_id="in_sequencing",
        received_at=datetime.now(),
        prepared_at=datetime.now(),
    )


@pytest.fixture
def order_with_cases(
    store: Store,
    helpers: StoreHelpers,
    order: Order,
    sample_in_preparation: Sample,
    sample_in_sequencing: Sample,
    sample_not_received: Sample,
) -> Order:
    case_not_received: Case = helpers.ensure_case(
        store=store,
        customer=order.customer,
        order=order,
        case_name="case_not_received",
        case_id="case_not_received",
    )
    case_in_preparation: Case = helpers.ensure_case(
        store=store,
        customer=order.customer,
        order=order,
        case_name="case_in_preparation",
        case_id="case_in_preparation",
    )
    case_in_sequencing: Case = helpers.ensure_case(
        store=store,
        customer=order.customer,
        order=order,
        case_name="case_in_sequencing",
        case_id="case_in_sequencing",
    )
    # Add samples to case that has not been received
    helpers.add_relationship(store=store, sample=sample_not_received, case=case_not_received)
    helpers.add_relationship(store=store, sample=sample_in_preparation, case=case_not_received)
    helpers.add_relationship(store=store, sample=sample_in_sequencing, case=case_not_received)

    # Add samples to case that is in preparation
    helpers.add_relationship(store=store, sample=sample_in_preparation, case=case_in_preparation)
    helpers.add_relationship(store=store, sample=sample_in_sequencing, case=case_in_preparation)

    # Add samples to case that is in sequencing
    helpers.add_relationship(store=store, sample=sample_in_sequencing, case=case_in_sequencing)
    return order


@pytest.fixture
def order_with_not_received_samples(
    store: Store,
    helpers: StoreHelpers,
    order: Order,
    sample_not_received: Sample,
    another_sample_not_received: Sample,
) -> Order:
    case_not_received: Case = helpers.ensure_case(
        store=store,
        customer=order.customer,
        order=order,
        case_name="case_not_received",
        case_id="case_not_received",
    )
    # Add samples to case that has not been received
    helpers.add_relationship(store=store, sample=sample_not_received, case=case_not_received)
    helpers.add_relationship(
        store=store, sample=another_sample_not_received, case=case_not_received
    )
    return order


@pytest.fixture
def order_with_two_cases(
    order: Order,
    sample_not_received: Sample,
    store: Store,
    helpers: StoreHelpers,
) -> Order:
    case_1: Case = helpers.ensure_case(
        store=store,
        customer=order.customer,
        order=order,
        case_name="case_not_received_1",
        case_id="case_not_received_1",
    )
    case_2: Case = helpers.ensure_case(
        store=store,
        customer=order.customer,
        order=order,
        case_name="case_not_received_2",
        case_id="case_not_received_2",
    )
    # Add samples
    helpers.add_relationship(store=store, sample=sample_not_received, case=case_1)
    helpers.add_relationship(store=store, sample=sample_not_received, case=case_2)
    return order



import logging
import re
import shutil
from datetime import datetime
from pathlib import Path
from typing import Any

import click

from cg.constants import EXIT_FAIL, EXIT_SUCCESS, Priority, Workflow
from cg.constants.constants import FileExtensions
from cg.constants.tb import AnalysisStatus
from cg.exc import CgDataError
from cg.meta.workflow.analysis import AnalysisAPI
from cg.meta.workflow.fastq import MicrosaltFastqHandler
from cg.meta.workflow.microsalt.quality_controller import QualityController
from cg.meta.workflow.microsalt.quality_controller.models import QualityResult
from cg.meta.workflow.microsalt.utils import get_most_recent_project_directory
from cg.models.cg_config import CGConfig
from cg.store.models import Case, Sample
from cg.utils import Process

LOG = logging.getLogger(__name__)


class MicrosaltAnalysisAPI(AnalysisAPI):
    """API to manage Microsalt Analyses"""

    def __init__(self, config: CGConfig, workflow: Workflow = Workflow.MICROSALT):
        super().__init__(workflow, config)
        self.root_dir = config.microsalt.root
        self.queries_path = config.microsalt.queries_path
        self.quality_checker = QualityController(config.status_db)

    @property
    def use_read_count_threshold(self) -> bool:
        return False

    @property
    def fastq_handler(self):
        return MicrosaltFastqHandler

    @property
    def conda_binary(self) -> str:
        return self.config.microsalt.conda_binary

    @property
    def process(self) -> Process:
        if not self._process:
            self._process = Process(
                binary=self.config.microsalt.binary_path,
                conda_binary=f"{self.conda_binary}" if self.conda_binary else None,
                environment=self.config.microsalt.conda_env,
            )
        return self._process

    def clean_run_dir(self, case_id: str, yes: bool, case_path: list[Path] | Path) -> int:
        """Remove workflow run directories for a MicroSALT case."""

        if not case_path:
            LOG.info(
                f"There is no case paths for case {case_id}. Setting cleaned at to {datetime.now()}"
            )
            self.clean_analyses(case_id=case_id)
            return EXIT_SUCCESS

        for analysis_path in case_path:
            if yes or click.confirm(
                f"Are you sure you want to remove all files in {analysis_path}?"
            ):
                if analysis_path.is_symlink():
                    LOG.warning(
                        f"Will not automatically delete symlink: {analysis_path}, delete it manually",
                    )
                    return EXIT_FAIL

                shutil.rmtree(analysis_path, ignore_errors=True)
                LOG.info(f"Cleaned {analysis_path}")

        self.clean_analyses(case_id=case_id)
        return EXIT_SUCCESS

    def get_case_fastq_path(self, case_id: str) -> Path:
        """Get fastq paths for a case."""
        return Path(self.root_dir, "fastq", case_id)

    def get_config_path(self, filename: str) -> Path:
        return Path(self.queries_path, filename).with_suffix(".json")

    def get_job_ids_path(self, case_id: str) -> Path:
        project_id: str = self.get_project_id(case_id)
        job_ids_path = Path(
            self.root_dir,
            "results",
            "reports",
            "trailblazer",
            f"{project_id}_slurm_ids{FileExtensions.YAML}",
        )
        # Necessary due to how microsalt structures its output
        self._ensure_old_job_ids_are_removed(job_ids_path)
        return job_ids_path

    def _ensure_old_job_ids_are_removed(self, job_ids_path: Path) -> None:
        is_yaml_file: bool = job_ids_path.suffix == FileExtensions.YAML
        if job_ids_path.exists() and is_yaml_file:
            job_ids_path.unlink()

    def get_deliverables_file_path(self, case_id: str) -> Path:
        """Returns a path where the microSALT deliverables file for the order_id should be
        located"""
        case_obj: Case = self.status_db.get_case_by_internal_id(internal_id=case_id)
        order_id: str = case_obj.name
        deliverables_file_path = Path(
            self.root_dir,
            "results",
            "reports",
            "deliverables",
            f"{order_id}_deliverables.yaml",
        )
        if deliverables_file_path.exists():
            LOG.info("Found deliverables file %s", deliverables_file_path)
        return deliverables_file_path

    def get_sample_fastq_destination_dir(self, case: Case, sample: Sample) -> Path:
        return Path(self.get_case_fastq_path(case_id=case.internal_id), sample.internal_id)

    def link_fastq_files(self, case_id: str, sample_id: str | None, dry_run: bool = False) -> None:
        case_obj: Case = self.status_db.get_case_by_internal_id(internal_id=case_id)
        samples: list[Sample] = self.get_samples(case_id=case_id, sample_id=sample_id)
        for sample_obj in samples:
            self.link_fastq_files_for_sample(case=case_obj, sample=sample_obj)

    def get_samples(self, case_id: str, sample_id: str | None = None) -> list[Sample]:
        """Returns a list of samples to configure
        If sample_id is specified, will return a list with only this sample_id.
        Otherwise, returns all samples in given case"""
        if sample_id:
            return [self.status_db.get_sample_by_internal_id(internal_id=sample_id)]

        case_obj: Case = self.status_db.get_case_by_internal_id(internal_id=case_id)
        return [link.sample for link in case_obj.links]

    def get_lims_comment(self, sample_id: str) -> str:
        """Returns the comment associated with a sample stored in lims"""
        comment: str = self.lims_api.get_sample_comment(sample_id) or ""
        return comment if re.match(r"\w{4}\d{2,3}", comment) else ""

    def get_organism(self, sample_obj: Sample) -> str:
        """Organism
        - Fallback based on reference, ‘Other species’ and ‘Comment’.
        Default to "Unset"."""

        if not sample_obj.organism:
            raise CgDataError("Organism missing on Sample")

        organism: str = sample_obj.organism.internal_id.strip()
        comment: str = self.get_lims_comment(sample_id=sample_obj.internal_id)

        if "gonorrhoeae" in organism:
            organism = "Neisseria spp."
        elif "Cutibacterium acnes" in organism:
            organism = "Propionibacterium acnes"

        if organism == "VRE":
            reference = sample_obj.organism.reference_genome
            has_comment = bool(comment)
            if reference == "NC_017960.1":
                organism = "Enterococcus faecium"
            elif reference == "NC_004668.1":
                organism = "Enterococcus faecalis"
            elif has_comment:
                organism = comment

        return organism

    def get_parameters(self, sample_obj: Sample) -> dict[str, str]:
        """Fill a dict with case config information for one sample"""

        sample_id = sample_obj.internal_id
        method_library_prep = self.lims_api.get_prep_method(sample_id)
        method_sequencing = self.lims_api.get_sequencing_method(sample_id)
        priority = (
            Priority.research.name if sample_obj.priority_int == 0 else Priority.standard.name
        )

        return {
            "CG_ID_project": self.get_project(sample_id),
            "Customer_ID_project": sample_obj.original_ticket,
            "CG_ID_sample": sample_obj.internal_id,
            "Customer_ID_sample": sample_obj.name,
            "organism": self.get_organism(sample_obj),
            "priority": priority,
            "reference": sample_obj.reference_genome,
            "Customer_ID": sample_obj.customer.internal_id,
            "application_tag": sample_obj.application_version.application.tag,
            "date_arrival": str(sample_obj.received_at or datetime.min),
            "date_sequencing": str(sample_obj.last_sequenced_at or datetime.min),
            "date_libprep": str(sample_obj.prepared_at or datetime.min),
            "method_libprep": method_library_prep or "Not in LIMS",
            "method_sequencing": method_sequencing or "Not in LIMS",
            "sequencing_qc_passed": sample_obj.sequencing_qc,
        }

    def get_project(self, sample_id: str) -> str:
        """Get LIMS project for a sample"""
        return self.lims_api.get_sample_project(sample_id)

    def resolve_case_sample_id(
        self, sample: bool, ticket: bool, unique_id: Any
    ) -> tuple[str, str | None]:
        """Resolve case_id and sample_id w based on input arguments."""
        if ticket and sample:
            LOG.error("Flags -t and -s are mutually exclusive!")
            raise click.Abort

        if ticket:
            case_id, sample_id = self.get_case_id_from_ticket(unique_id)

        elif sample:
            case_id, sample_id = self.get_case_id_from_sample(unique_id)

        else:
            case_id, sample_id = self.get_case_id_from_case(unique_id)

        return case_id, sample_id

    def get_case_id_from_ticket(self, unique_id: str) -> tuple[str, None]:
        """If ticked is provided as argument, finds the corresponding case_id and returns it.
        Since sample_id is not specified, nothing is returned as sample_id"""
        case: Case = self.status_db.get_case_by_name(name=unique_id)
        if not case:
            LOG.error(f"No case found for ticket number:  {unique_id}")
            raise click.Abort
        case_id = case.internal_id
        return case_id, None

    def get_case_id_from_sample(self, unique_id: str) -> tuple[str, str]:
        """If sample is specified, finds the corresponding case_id to which this sample belongs.
        The case_id is to be used for identifying the appropriate path to link fastq files and store the analysis output
        """
        sample: Sample = self.status_db.get_sample_by_internal_id(unique_id)
        if not sample:
            LOG.error(f"No sample found with id: {unique_id}")
            raise click.Abort
        case_id = sample.links[0].case.internal_id
        sample_id = sample.internal_id
        return case_id, sample_id

    def get_case_id_from_case(self, unique_id: str) -> tuple[str, None]:
        """If case_id is specified, validates the presence of case_id in database and returns it"""
        case_obj: Case = self.status_db.get_case_by_internal_id(unique_id)
        if not case_obj:
            LOG.error(f"No case found with the id:  {unique_id}")
            raise click.Abort
        case_id = case_obj.internal_id
        return case_id, None

    def get_cases_to_store(self) -> list[Case]:
        cases_qc_ready: list[Case] = self.get_completed_cases()
        cases_to_store: list[Case] = []
        LOG.info(f"Found {len(cases_qc_ready)} cases to perform QC on!")

        for case in cases_qc_ready:
            case_run_dir: Path | None = self.get_case_path(case.internal_id)
            LOG.info(f"Checking QC for case {case.internal_id} in {case_run_dir}")

            if self.quality_checker.is_qc_required(case_run_dir):
                LOG.info(f"QC required for case {case.internal_id}")
                metrics_file_path = self.get_metrics_file_path(case.internal_id)

                if not metrics_file_path.exists():
                    continue

                result: QualityResult = self.quality_checker.quality_control(metrics_file_path)
                self.trailblazer_api.add_comment(case_id=case.internal_id, comment=result.summary)
                if result.passes_qc:
                    cases_to_store.append(case)
                else:
                    self.trailblazer_api.set_analysis_status(
                        case_id=case.internal_id, status=AnalysisStatus.FAILED
                    )
            else:
                cases_to_store.append(case)

        return cases_to_store

    def get_completed_cases(self) -> list[Case]:
        """Return cases that are completed in trailblazer."""
        return [
            case
            for case in self.status_db.get_running_cases_in_workflow(self.workflow)
            if self.trailblazer_api.is_latest_analysis_completed(case.internal_id)
        ]

    def get_metrics_file_path(self, case_id: str) -> Path:
        """Return path to metrics file for a case."""
        project_id: str = self.get_project_id(case_id)
        case_run_dir: Path = self.get_case_path(case_id)
        return Path(case_run_dir, f"{project_id}{FileExtensions.JSON}")

    def extract_project_id(self, sample_id: str) -> str:
        return sample_id.rsplit("A", maxsplit=1)[0]

    def get_project_id(self, case_id: str) -> str:
        case: Case = self.status_db.get_case_by_internal_id(case_id)
        sample_id: str = case.links[0].sample.internal_id
        return self.extract_project_id(sample_id)

    def get_results_dir(self) -> Path:
        return Path(self.root_dir, "results")

    def get_case_path(self, case_id: str) -> Path:
        project_id: str = self.get_project_id(case_id)
        results_dir: Path = self.get_results_dir()
        return get_most_recent_project_directory(project_id=project_id, directory=results_dir)



from cg.store.models import Case


def get_fastq_delivery_path(case: Case) -> str:
    ticket_id: str = case.latest_ticket
    customer_id: str = case.customer.internal_id
    return f"/home/{customer_id}/inbox/{ticket_id}"


def get_scout_link(case: Case) -> str:
    customer_id: str = case.customer.internal_id
    case_name: str = case.name
    return f"https://scout.scilifelab.se/{customer_id}/{case_name}"


def get_pangolin_delivery_path(case: Case) -> str:
    customer_id: str = case.customer.internal_id
    return f"/home/{customer_id}/inbox/wwLab_automatisk_hamtning"


def get_batch_id_for_case(case: Case) -> str:
    return case.name.split("-")[1]


def get_statina_link(batch_id: str) -> str:
    return f"https://statina.clinicalgenomics.se/batches/{batch_id}"



import logging

import click

from cg.cli.workflow.commands import link, resolve_compression, store, store_available
from cg.constants import EXIT_FAIL, EXIT_SUCCESS
from cg.exc import AnalysisNotReadyError, CgError
from cg.meta.workflow.analysis import AnalysisAPI
from cg.meta.workflow.fluffy import FluffyAnalysisAPI
from cg.models.cg_config import CGConfig

OPTION_DRY = click.option(
    "-d", "--dry-run", "dry_run", help="Print command to console without executing", is_flag=True
)
ARGUMENT_CASE_ID = click.argument("case_id", required=True)
OPTION_EXTERNAL_REF = click.option("-e", "--external-ref", is_flag=True)

LOG = logging.getLogger(__name__)


@click.group(invoke_without_command=True)
@click.pass_context
def fluffy(context: click.Context):
    """
    Fluffy workflow
    """
    AnalysisAPI.get_help(context)
    context.obj.meta_apis["analysis_api"] = FluffyAnalysisAPI(config=context.obj)


fluffy.add_command(link)
fluffy.add_command(resolve_compression)
fluffy.add_command(store)
fluffy.add_command(store_available)


@fluffy.command("create-samplesheet")
@ARGUMENT_CASE_ID
@OPTION_DRY
@click.pass_obj
def create_samplesheet(context: CGConfig, case_id: str, dry_run: bool):
    """
    Write modified samplesheet file to case folder
    """
    analysis_api: FluffyAnalysisAPI = context.meta_apis["analysis_api"]
    analysis_api.status_db.verify_case_exists(case_internal_id=case_id)
    analysis_api.make_sample_sheet(case_id=case_id, dry_run=dry_run)


@fluffy.command()
@ARGUMENT_CASE_ID
@OPTION_DRY
@click.option("-c", "--config", help="Path to fluffy config in .json format")
@OPTION_EXTERNAL_REF
@click.pass_obj
def run(context: CGConfig, case_id: str, dry_run: bool, config: str, external_ref: bool = False):
    """
    Run Fluffy analysis
    """
    analysis_api: FluffyAnalysisAPI = context.meta_apis["analysis_api"]
    analysis_api.status_db.verify_case_exists(case_internal_id=case_id)
    analysis_api.run_fluffy(
        case_id=case_id, workflow_config=config, dry_run=dry_run, external_ref=external_ref
    )
    if dry_run:
        return
    # Submit analysis for tracking in Trailblazer
    try:
        analysis_api.add_pending_trailblazer_analysis(case_id=case_id)
        LOG.info(f"Submitted case {case_id} to Trailblazer!")
    except Exception as error:
        LOG.warning(f"Unable to submit job file to Trailblazer, raised error: {error}")

    analysis_api.set_statusdb_action(case_id=case_id, action="running")


@fluffy.command()
@ARGUMENT_CASE_ID
@OPTION_DRY
@click.option("-c", "--config", help="Path to fluffy config in .json format")
@OPTION_EXTERNAL_REF
@click.pass_context
def start(
    context: click.Context,
    case_id: str,
    dry_run: bool,
    external_ref: bool = False,
    config: str = None,
):
    """
    Starts full Fluffy analysis workflow
    """
    LOG.info(f"Starting full Fluffy workflow for {case_id}")
    if dry_run:
        LOG.info("Dry run: the executed commands will not produce output!")
    analysis_api: FluffyAnalysisAPI = context.obj.meta_apis["analysis_api"]
    analysis_api.prepare_fastq_files(case_id=case_id, dry_run=dry_run)
    context.invoke(link, case_id=case_id, dry_run=dry_run)
    context.invoke(create_samplesheet, case_id=case_id, dry_run=dry_run)
    context.invoke(run, case_id=case_id, config=config, dry_run=dry_run, external_ref=external_ref)


@fluffy.command("start-available")
@OPTION_DRY
@click.pass_context
def start_available(context: click.Context, dry_run: bool = False):
    """Start full analysis workflow for all cases ready for analysis"""

    analysis_api: FluffyAnalysisAPI = context.obj.meta_apis["analysis_api"]

    exit_code: int = EXIT_SUCCESS
    for case_obj in analysis_api.get_cases_to_analyze():
        try:
            context.invoke(start, case_id=case_obj.internal_id, dry_run=dry_run)
        except AnalysisNotReadyError as error:
            LOG.error(error)
        except CgError as error:
            LOG.error(error)
            exit_code = EXIT_FAIL
        except Exception as error:
            LOG.error(f"Unspecified error occurred: {error}")
            exit_code = EXIT_FAIL
    if exit_code:
        raise click.Abort



"""Module for nf-core analysis API tests."""

import logging
from typing import Any

import pytest
from _pytest.fixtures import FixtureRequest
from _pytest.logging import LogCaptureFixture

from cg.constants import Workflow
from cg.meta.workflow.nf_analysis import NfAnalysisAPI
from cg.models.cg_config import CGConfig
from tests.cli.workflow.conftest import deliverables_template_content


@pytest.mark.parametrize(
    "workflow",
    [Workflow.RNAFUSION, Workflow.TAXPROFILER, Workflow.TOMTE],
)
def test_create_metrics_deliverables_content(
    workflow: Workflow,
    caplog: LogCaptureFixture,
    request: FixtureRequest,
):
    """Test metrics deliverables file content function for Taxprofiler and Rnafusion."""
    caplog.set_level(logging.INFO)

    # GIVEN each fixture is being initialised
    context: CGConfig = request.getfixturevalue(f"{workflow}_context")

    metrics_deliverables: dict = request.getfixturevalue(f"{workflow}_metrics_deliverables")
    case_id: str = request.getfixturevalue(f"{workflow}_case_id")
    request.getfixturevalue(f"{workflow}_mock_analysis_finish")

    # GIVEN a Nextflow workflow analysis API and a list of QC metrics
    analysis_api: NfAnalysisAPI = context.meta_apis["analysis_api"]

    # WHEN writing the metrics deliverables file
    metrics_deliverables_content: dict[str, list[dict[str, Any]]] = (
        analysis_api.create_metrics_deliverables_content(case_id)
    )

    # THEN assert that the content created is correct
    assert metrics_deliverables_content == metrics_deliverables


@pytest.mark.parametrize(
    "workflow",
    [Workflow.RNAFUSION, Workflow.TAXPROFILER, Workflow.TOMTE],
)
def test_get_formatted_file_deliverable(
    workflow: Workflow,
    sample_id: str,
    sample_name: str,
    caplog: LogCaptureFixture,
    deliverables_template_content,
    request: FixtureRequest,
):
    """Test the formatted file deliverable with the case and sample attributes for Taxprofiler and Rnafusion."""

    caplog.set_level(logging.INFO)

    # GIVEN each fixture is being initialised
    context: CGConfig = request.getfixturevalue(f"{workflow}_context")
    case_id: str = request.getfixturevalue(f"{workflow}_case_id")

    caplog.set_level(logging.INFO)
    analysis_api: NfAnalysisAPI = context.meta_apis["analysis_api"]

    # WHEN formatting the template
    for field in deliverables_template_content:
        formatted_file_deliverable = analysis_api.get_formatted_file_deliverable(
            field,
            case_id,
            sample_id,
            sample_name,
            "get_case_path",
        )

        # THEN assert that the new path is correct
        assert "get_case_path" in formatted_file_deliverable.path



"""Module for Balsamic Analysis API."""

import logging
from pathlib import Path

from housekeeper.store.models import File, Version
from pydantic.v1 import EmailStr, ValidationError

from cg.constants import Workflow
from cg.constants.constants import FileFormat, SampleType
from cg.constants.housekeeper_tags import BalsamicAnalysisTag
from cg.constants.observations import ObservationsFileWildcards
from cg.constants.priority import SlurmQos
from cg.constants.sequencing import Variants
from cg.constants.subject import Sex
from cg.exc import BalsamicStartError, CgError
from cg.io.controller import ReadFile
from cg.meta.workflow.analysis import AnalysisAPI
from cg.meta.workflow.fastq import BalsamicFastqHandler
from cg.models.balsamic.analysis import BalsamicAnalysis
from cg.models.balsamic.metrics import (
    BalsamicMetricsBase,
    BalsamicTargetedQCMetrics,
    BalsamicWGSQCMetrics,
)
from cg.models.cg_config import CGConfig
from cg.store.models import Case, CaseSample, Sample
from cg.utils import Process
from cg.utils.utils import build_command_from_dict, get_string_from_list_by_pattern

LOG = logging.getLogger(__name__)


class BalsamicAnalysisAPI(AnalysisAPI):
    """Handles communication between BALSAMIC processes
    and the rest of CG infrastructure"""

    __BALSAMIC_APPLICATIONS = {"wgs", "wes", "tgs"}
    __BALSAMIC_BED_APPLICATIONS = {"wes", "tgs"}

    def __init__(
        self,
        config: CGConfig,
        workflow: Workflow = Workflow.BALSAMIC,
    ):
        super().__init__(workflow=workflow, config=config)
        self.account: str = config.balsamic.slurm.account
        self.binary_path: str = config.balsamic.binary_path
        self.balsamic_cache: str = config.balsamic.balsamic_cache
        self.conda_binary: str = config.balsamic.conda_binary
        self.conda_env: str = config.balsamic.conda_env
        self.bed_path: str = config.balsamic.bed_path
        self.cadd_path: str = config.balsamic.cadd_path
        self.genome_interval_path: str = config.balsamic.genome_interval_path
        self.gnomad_af5_path: str = config.balsamic.gnomad_af5_path
        self.gens_coverage_female_path: str = config.balsamic.gens_coverage_female_path
        self.gens_coverage_male_path: str = config.balsamic.gens_coverage_male_path
        self.email: EmailStr = config.balsamic.slurm.mail_user
        self.loqusdb_path: str = config.balsamic.loqusdb_path
        self.pon_path: str = config.balsamic.pon_path
        self.qos: SlurmQos = config.balsamic.slurm.qos
        self.root_dir: str = config.balsamic.root
        self.swegen_path: str = config.balsamic.swegen_path

    @property
    def root(self) -> str:
        return self.root_dir

    @property
    def use_read_count_threshold(self) -> bool:
        return True

    @property
    def fastq_handler(self):
        return BalsamicFastqHandler

    @property
    def process(self):
        if not self._process:
            self._process = Process(
                binary=self.binary_path, conda_binary=self.conda_binary, environment=self.conda_env
            )
        return self._process

    @property
    def PON_file_suffix(self) -> str:
        """Panel of normals reference file suffix (<panel-bed>_<PON>_<version>.cnn)"""

        return "CNVkit_PON_reference_v*.cnn"

    def get_case_path(self, case_id: str) -> Path:
        """Returns a path where the Balsamic case for the case_id should be located"""
        return Path(self.root_dir, case_id)

    def get_cases_to_analyze(self) -> list[Case]:
        cases_query: list[Case] = self.status_db.cases_to_analyze(
            workflow=self.workflow, threshold=self.use_read_count_threshold
        )
        cases_to_analyze = []
        for case_obj in cases_query:
            if case_obj.action == "analyze" or not case_obj.latest_analyzed:
                cases_to_analyze.append(case_obj)
            elif (
                self.trailblazer_api.get_latest_analysis_status(case_id=case_obj.internal_id)
                == "failed"
            ):
                cases_to_analyze.append(case_obj)
        return cases_to_analyze

    def get_deliverables_file_path(self, case_id: str) -> Path:
        """Returns a path where the Balsamic deliverables file for the case_id should be located.

        (Optional) Checks if deliverables file exists
        """
        return Path(
            self.root_dir,
            case_id,
            "analysis",
            "delivery_report",
            case_id + ".hk",
        )

    def get_case_config_path(self, case_id: str) -> Path:
        """Generates a path where the Balsamic config for the case_id should be located.

        (Optional) Checks if config file exists.
        """
        return Path(self.root_dir, case_id, case_id + ".json")

    def get_job_ids_path(self, case_id: str) -> Path:
        return Path(self.root_dir, case_id, "analysis", "slurm_jobids.yaml")

    def get_bundle_deliverables_type(self, case_id: str) -> str:
        """Return the analysis type for a case

        Analysis types are any of ["tumor_wgs", "tumor_normal_wgs", "tumor_panel", "tumor_normal_panel"]
        """
        LOG.debug("Fetch analysis type for %s", case_id)
        number_of_samples: int = len(
            self.status_db.get_case_by_internal_id(internal_id=case_id).links
        )

        application_type: str = self.get_application_type(
            self.status_db.get_case_by_internal_id(internal_id=case_id).links[0].sample
        )
        sample_type = "tumor"
        if number_of_samples == 2:
            sample_type = "tumor_normal"
        if application_type != "wgs":
            application_type = "panel"
        analysis_type = "_".join([sample_type, application_type])
        LOG.info("Found analysis type %s", analysis_type)
        return analysis_type

    def get_sample_fastq_destination_dir(self, case: Case, sample: Sample = None) -> Path:
        """Return the path to the FASTQ destination directory."""
        return Path(self.get_case_path(case.internal_id), FileFormat.FASTQ)

    def link_fastq_files(self, case_id: str, dry_run: bool = False) -> None:
        """Link fastq files from Housekeeper to Balsamic case working directory."""
        case: Case = self.status_db.get_case_by_internal_id(internal_id=case_id)
        for link in case.links:
            self.link_fastq_files_for_sample(case=case, sample=link.sample)

    @staticmethod
    def get_sample_type(sample_obj: Sample) -> SampleType:
        """Return the tissue type of the sample."""
        if sample_obj.is_tumour:
            return SampleType.TUMOR
        return SampleType.NORMAL

    def get_derived_bed(self, panel_bed: str) -> Path | None:
        """Returns the verified capture kit path or the derived panel BED path."""
        if not panel_bed:
            return None
        panel_bed: Path = Path(panel_bed)
        if panel_bed.is_file():
            return panel_bed
        derived_panel_bed: Path = Path(
            self.bed_path,
            self.status_db.get_bed_version_by_short_name(
                bed_version_short_name=panel_bed.as_posix()
            ).filename,
        )
        if not derived_panel_bed.is_file():
            raise BalsamicStartError(
                f"{panel_bed} or {derived_panel_bed} are not valid paths to a BED file. "
                f"Please provide absolute path to desired BED file or a valid bed shortname!"
            )
        return derived_panel_bed

    def get_verified_bed(self, panel_bed: str, sample_data: dict) -> str | None:
        """Takes a dict with samples and attributes.
        Retrieves unique attributes for application type and target_bed.
        Verifies that those attributes are the same across multiple samples,
        where applicable.
        Verifies that the attributes are valid BALSAMIC attributes
        If application type requires bed, returns path to bed.

        Raises BalsamicStartError:
        - When application type invalid for balsamic
        - When multiple samples have different parameters
        - When bed file required for analysis, but is not set or cannot be retrieved.
        """

        panel_bed: Path | None = self.get_derived_bed(panel_bed)
        application_types = {v["application_type"].lower() for k, v in sample_data.items()}
        target_beds: set = {v["target_bed"] for k, v in sample_data.items()}

        if not application_types.issubset(self.__BALSAMIC_APPLICATIONS):
            raise BalsamicStartError("Case application not compatible with BALSAMIC")
        if len(application_types) != 1:
            raise BalsamicStartError("Multiple application types found in LIMS")
        if not application_types.issubset(self.__BALSAMIC_BED_APPLICATIONS):
            if panel_bed:
                raise BalsamicStartError("Cannot set panel_bed for WGS sample!")
            return None
        if panel_bed:
            return panel_bed.as_posix()
        if len(target_beds) == 1:
            target_bed = target_beds.pop()
            if not target_bed:
                raise BalsamicStartError(
                    f"Application type {application_types.pop()} requires a bed file to be analyzed!"
                )
            return Path(self.bed_path, target_bed).as_posix()

    def get_verified_pon(self, panel_bed: str, pon_cnn: str) -> str | None:
        """Returns the validated PON or extracts the latest one available if it is not provided

        Raises BalsamicStartError:
            When there is a missmatch between the PON and the panel bed file names
        """

        if pon_cnn:
            latest_pon = pon_cnn
            if Path(panel_bed).stem not in Path(latest_pon).stem:
                raise BalsamicStartError(
                    f"The specified PON reference file {latest_pon} does not match the panel bed {panel_bed}"
                )
        else:
            latest_pon = self.get_latest_pon_file(panel_bed)
            if latest_pon:
                LOG.info(
                    f"The following PON reference file will be used for the analysis: {latest_pon}"
                )

        return latest_pon

    def get_latest_pon_file(self, panel_bed: str) -> str | None:
        """Returns the latest PON cnn file associated to a specific capture bed"""

        if not panel_bed:
            raise BalsamicStartError("BALSAMIC PON workflow requires a panel bed to be specified")

        pon_list = Path(self.pon_path).glob(f"*{Path(panel_bed).stem}_{self.PON_file_suffix}")
        sorted_pon_files = sorted(
            pon_list,
            key=lambda file: int(file.stem.split("_v")[-1]),
            reverse=True,
        )

        return sorted_pon_files[0].as_posix() if sorted_pon_files else None

    @staticmethod
    def get_verified_sex(sample_data: dict) -> Sex:
        """Takes a dict with samples and attributes, and returns a verified case sex provided by the customer."""
        sex: Sex = next(iter(sample_data.values()))["sex"]
        if all(val["sex"] == sex for val in sample_data.values()) and sex in set(
            value for value in Sex
        ):
            if sex not in [Sex.FEMALE, Sex.MALE]:
                LOG.warning(f"The provided sex is unknown, setting {Sex.FEMALE} as the default")
                sex = Sex.FEMALE

            return sex
        else:
            LOG.error(f"Unable to retrieve a valid sex from samples: {sample_data.keys()}")
            raise BalsamicStartError

    def get_verified_samples(self, case_id: str) -> dict[str, str]:
        """Return a verified tumor and normal sample dictionary."""
        tumor_samples: list[Sample] = self.status_db.get_samples_by_type(
            case_id=case_id, sample_type=SampleType.TUMOR
        )
        normal_samples: list[Sample] = self.status_db.get_samples_by_type(
            case_id=case_id, sample_type=SampleType.NORMAL
        )
        if (
            not tumor_samples
            and not normal_samples
            or len(tumor_samples) > 1
            or len(normal_samples) > 1
        ):
            LOG.error(f"Case {case_id} has an invalid number of samples")
            raise BalsamicStartError

        tumor_sample_id: str = tumor_samples[0].internal_id if tumor_samples else None
        normal_sample_id: str = normal_samples[0].internal_id if normal_samples else None
        if normal_sample_id and not tumor_sample_id:
            LOG.warning(
                f"Only a normal sample was found for case {case_id}. "
                f"Balsamic analysis will treat it as a tumor sample."
            )
            tumor_sample_id, normal_sample_id = normal_sample_id, None
        return {"tumor_sample_name": tumor_sample_id, "normal_sample_name": normal_sample_id}

    def get_latest_raw_file_data(self, case_id: str, tags: list) -> dict | list:
        """Retrieves the data of the latest file associated to a specific case ID and a list of tags."""

        version: Version = self.housekeeper_api.last_version(bundle=case_id)
        raw_file: File = self.housekeeper_api.get_latest_file(
            bundle=case_id, version=version.id, tags=tags
        )

        if not raw_file:
            raise FileNotFoundError(
                f"No file associated to {tags} was found in housekeeper for {case_id}"
            )
        return ReadFile.get_content_from_file(
            file_format=FileFormat.YAML, file_path=Path(raw_file.full_path)
        )

    def get_latest_metadata(self, case_id: str) -> BalsamicAnalysis:
        """Get the latest metadata of a specific BALSAMIC case"""

        config_raw_data = self.get_latest_raw_file_data(case_id, BalsamicAnalysisTag.CONFIG)
        metrics_raw_data = self.get_latest_raw_file_data(case_id, BalsamicAnalysisTag.QC_METRICS)

        if config_raw_data and metrics_raw_data:
            try:
                balsamic_analysis = self.parse_analysis(
                    config_raw=config_raw_data, qc_metrics_raw=metrics_raw_data
                )
                return balsamic_analysis
            except ValidationError as error:
                LOG.error(
                    "get_latest_metadata failed for '%s', missing attribute: %s",
                    case_id,
                    error,
                )
                raise error
        else:
            LOG.error(f"Unable to retrieve the latest metadata for {case_id}")
            raise CgError

    def parse_analysis(self, config_raw: dict, qc_metrics_raw: dict, **kwargs) -> BalsamicAnalysis:
        """Returns a formatted BalsamicAnalysis object"""

        sequencing_type = config_raw["analysis"]["sequencing_type"]
        qc_metrics = dict()

        for value in qc_metrics_raw:
            sample_metric = BalsamicMetricsBase(**value)
            try:
                qc_metrics[sample_metric.id].update(
                    {sample_metric.name.lower(): sample_metric.value}
                )
            except KeyError:
                qc_metrics[sample_metric.id] = {sample_metric.name.lower(): sample_metric.value}

        return BalsamicAnalysis(
            config=config_raw,
            sample_metrics=self.cast_metrics_type(sequencing_type, qc_metrics),
        )

    @staticmethod
    def cast_metrics_type(
        sequencing_type: str, metrics: dict
    ) -> BalsamicTargetedQCMetrics | BalsamicWGSQCMetrics:
        """Cast metrics model type according to the sequencing type"""

        if metrics:
            for k, v in metrics.items():
                metrics[k] = (
                    BalsamicWGSQCMetrics(**v)
                    if sequencing_type == "wgs"
                    else BalsamicTargetedQCMetrics(**v)
                )

        return metrics

    @staticmethod
    def get_latest_file_by_pattern(directory: Path, pattern: str) -> str | None:
        """Returns the latest file (<file_name>-<date>-.vcf.gz) matching a pattern from a specific directory."""
        available_files: iter = sorted(
            Path(directory).glob(f"*{pattern}*.vcf.gz"),
            key=lambda file: file.stem.split("-"),
            reverse=True,
        )
        return str(available_files[0]) if available_files else None

    def get_parsed_observation_file_paths(self, observations: list[str]) -> dict:
        """Returns a verified {option: path} observations dictionary."""
        verified_observations: dict[str, str] = {}
        for wildcard in list(ObservationsFileWildcards):
            file_path: str = get_string_from_list_by_pattern(observations, wildcard)
            verified_observations.update(
                {
                    wildcard: (
                        file_path
                        if file_path
                        else self.get_latest_file_by_pattern(
                            directory=self.loqusdb_path, pattern=wildcard
                        )
                    )
                }
            )

        return verified_observations

    def get_verified_gens_file_paths(self, sex: Sex) -> dict[str, str] | None:
        """Return a list of file path arguments for Gens."""
        return {
            "genome_interval": self.genome_interval_path,
            "gnomad_min_af5": self.gnomad_af5_path,
            "gens_coverage_pon": (
                self.gens_coverage_female_path
                if sex == Sex.FEMALE
                else self.gens_coverage_male_path
            ),
        }

    def get_swegen_verified_path(self, variants: Variants) -> str | None:
        """Return verified SweGen path."""
        swegen_file: str = self.get_latest_file_by_pattern(
            directory=self.swegen_path, pattern=variants
        )
        return swegen_file

    def get_verified_config_case_arguments(
        self,
        case_id: str,
        genome_version: str,
        panel_bed: str,
        pon_cnn: str,
        observations: list[str] = None,
        sex: str | None = None,
    ) -> dict:
        """Takes a dictionary with per-sample parameters,
        validates them, and transforms into command line arguments
        Raises BalsamicStartError:
            When no samples associated with case are marked for BALSAMIC analysis
        """
        sample_data = self.get_sample_params(case_id=case_id, panel_bed=panel_bed)
        if len(sample_data) == 0:
            raise BalsamicStartError(f"{case_id} has no samples tagged for BALSAMIC analysis!")

        verified_panel_bed = self.get_verified_bed(panel_bed=panel_bed, sample_data=sample_data)
        verified_pon = (
            self.get_verified_pon(pon_cnn=pon_cnn, panel_bed=verified_panel_bed)
            if verified_panel_bed
            else None
        )
        verified_sex: Sex = sex or self.get_verified_sex(sample_data=sample_data)

        verified_exome_argument: bool = self.has_case_only_exome_samples(case_id=case_id)

        config_case: dict[str, str] = {
            "case_id": case_id,
            "analysis_workflow": self.workflow,
            "genome_version": genome_version,
            "sex": verified_sex,
            "panel_bed": verified_panel_bed,
            "pon_cnn": verified_pon,
            "swegen_snv": self.get_swegen_verified_path(Variants.SNV),
            "swegen_sv": self.get_swegen_verified_path(Variants.SV),
            "exome": verified_exome_argument,
        }

        config_case.update(self.get_verified_samples(case_id=case_id))
        config_case.update(self.get_parsed_observation_file_paths(observations))
        (
            config_case.update(self.get_verified_gens_file_paths(sex=verified_sex))
            if not verified_panel_bed
            else None
        )

        return config_case

    @staticmethod
    def print_sample_params(case_id: str, sample_data: dict) -> None:
        """Outputs a table of samples to be displayed in log"""

        LOG.info(f"Case {case_id} has following BALSAMIC samples:")
        LOG.info(
            "{:<20} {:<20} {:<20} {:<20}".format(
                "SAMPLE ID", "TISSUE TYPE", "APPLICATION", "BED VERSION"
            )
        )
        for key in sample_data:
            LOG.info(
                "{:<20} {:<20} {:<20} {:<20}".format(
                    key,
                    str(sample_data[key]["tissue_type"]),
                    str(sample_data[key]["application_type"]),
                    str(sample_data[key]["target_bed"]),
                )
            )
        LOG.info("")

    def get_sample_params(self, case_id: str, panel_bed: str | None) -> dict:
        """Returns a dictionary of attributes for each sample in given family,
        where SAMPLE ID is used as key"""

        sample_data = {
            link_object.sample.internal_id: {
                "sex": link_object.sample.sex,
                "tissue_type": self.get_sample_type(link_object.sample).value,
                "application_type": self.get_application_type(link_object.sample),
                "target_bed": self.resolve_target_bed(panel_bed=panel_bed, link_object=link_object),
            }
            for link_object in self.status_db.get_case_by_internal_id(internal_id=case_id).links
        }

        self.print_sample_params(case_id=case_id, sample_data=sample_data)
        return sample_data

    def resolve_target_bed(self, panel_bed: str | None, link_object: CaseSample) -> str | None:
        if panel_bed:
            return panel_bed
        if self.get_application_type(link_object.sample) not in self.__BALSAMIC_BED_APPLICATIONS:
            return None
        return self.get_target_bed_from_lims(link_object.case.internal_id)

    def get_workflow_version(self, case_id: str) -> str:
        LOG.debug("Fetch workflow version")
        config_data: dict = ReadFile.get_content_from_file(
            file_format=FileFormat.JSON, file_path=self.get_case_config_path(case_id=case_id)
        )
        return config_data["analysis"]["BALSAMIC_version"]

    def config_case(
        self,
        case_id: str,
        gender: str,
        genome_version: str,
        panel_bed: str,
        pon_cnn: str,
        observations: list[str],
        cache_version: str,
        dry_run: bool = False,
    ) -> None:
        """Create config file for BALSAMIC analysis"""
        arguments = self.get_verified_config_case_arguments(
            case_id=case_id,
            sex=gender,
            genome_version=genome_version,
            panel_bed=panel_bed,
            pon_cnn=pon_cnn,
            observations=observations,
        )
        command = ["config", "case"]
        options = build_command_from_dict(
            {
                "--analysis-dir": self.root_dir,
                "--analysis-workflow": arguments.get("analysis_workflow"),
                "--balsamic-cache": self.balsamic_cache,
                "--cache-version": cache_version,
                "--cadd-annotations": self.cadd_path,
                "--cancer-germline-snv-observations": arguments.get("cancer_germline_snv"),
                "--cancer-germline-sv-observations": arguments.get("cancer_germline_sv"),
                "--cancer-somatic-snv-observations": arguments.get("cancer_somatic_snv"),
                "--cancer-somatic-sv-observations": arguments.get("cancer_somatic_sv"),
                "--case-id": arguments.get("case_id"),
                "--clinical-snv-observations": arguments.get("clinical_snv"),
                "--clinical-sv-observations": arguments.get("clinical_sv"),
                "--fastq-path": self.get_sample_fastq_destination_dir(
                    self.status_db.get_case_by_internal_id(case_id)
                ),
                "--gender": arguments.get("sex"),
                "--genome-interval": arguments.get("genome_interval"),
                "--genome-version": arguments.get("genome_version"),
                "--gens-coverage-pon": arguments.get("gens_coverage_pon"),
                "--gnomad-min-af5": arguments.get("gnomad_min_af5"),
                "--normal-sample-name": arguments.get("normal_sample_name"),
                "--panel-bed": arguments.get("panel_bed"),
                "--exome": arguments.get("exome"),
                "--pon-cnn": arguments.get("pon_cnn"),
                "--swegen-snv": arguments.get("swegen_snv"),
                "--swegen-sv": arguments.get("swegen_sv"),
                "--tumor-sample-name": arguments.get("tumor_sample_name"),
                "--umi-trim-length": arguments.get("umi_trim_length"),
            },
            exclude_true=True,
        )
        parameters = command + options
        self.process.run_command(parameters=parameters, dry_run=dry_run)

    def run_analysis(
        self,
        case_id: str,
        cluster_config: Path | None = None,
        slurm_quality_of_service: str | None = None,
        dry_run: bool = False,
    ) -> None:
        """Execute BALSAMIC run analysis with given options"""

        command = ["run", "analysis"]
        run_analysis = ["--run-analysis"] if not dry_run else []
        benchmark = ["--benchmark"]
        options = build_command_from_dict(
            {
                "--account": self.account,
                "--mail-user": self.email,
                "--qos": slurm_quality_of_service or self.get_slurm_qos_for_case(case_id=case_id),
                "--sample-config": self.get_case_config_path(case_id=case_id),
                "--cluster-config": cluster_config,
            }
        )
        parameters = command + options + run_analysis + benchmark
        self.process.run_command(parameters=parameters, dry_run=dry_run)

    def report_deliver(self, case_id: str, dry_run: bool = False) -> None:
        """Execute BALSAMIC report deliver with given options"""

        command = ["report", "deliver"]
        options = build_command_from_dict(
            {
                "--sample-config": self.get_case_config_path(case_id=case_id),
            }
        )
        parameters = command + options
        self.process.run_command(parameters=parameters, dry_run=dry_run)



"""Tests CLI common methods to create the case config for NF analyses."""

import logging
from pathlib import Path

import pytest
from _pytest.fixtures import FixtureRequest
from _pytest.logging import LogCaptureFixture
from click.testing import CliRunner

from cg.cli.workflow.base import workflow as workflow_cli
from cg.constants import EXIT_SUCCESS, Workflow
from cg.io.yaml import read_yaml
from cg.models.cg_config import CGConfig

LOG = logging.getLogger(__name__)


@pytest.mark.parametrize(
    "workflow",
    [Workflow.RNAFUSION, Workflow.TAXPROFILER, Workflow.TOMTE],
)
def test_metrics_deliver_without_options(
    cli_runner: CliRunner, workflow: Workflow, request: FixtureRequest
):
    """Test metrics-deliver for workflow without options."""
    context: CGConfig = request.getfixturevalue(f"{workflow}_context")

    # WHEN invoking the command without additional parameters
    result = cli_runner.invoke(workflow_cli, [workflow, "metrics-deliver"], obj=context)

    # THEN command should not exit successfully
    assert result.exit_code != EXIT_SUCCESS

    # THEN command log should inform about missing arguments
    assert "Missing argument" in result.output


@pytest.mark.parametrize(
    "workflow",
    [Workflow.RNAFUSION, Workflow.TAXPROFILER, Workflow.TOMTE],
)
def test_metrics_deliver_with_missing_case(
    cli_runner: CliRunner,
    caplog: LogCaptureFixture,
    case_id_does_not_exist: str,
    workflow: Workflow,
    request: FixtureRequest,
):
    """Test metrics-deliver for workflow with a missing case."""
    caplog.set_level(logging.ERROR)
    context: CGConfig = request.getfixturevalue(f"{workflow}_context")

    # GIVEN a case not in the StatusDB database
    assert not context.status_db.get_case_by_internal_id(internal_id=case_id_does_not_exist)

    # WHEN running
    result = cli_runner.invoke(
        workflow_cli, [workflow, "metrics-deliver", case_id_does_not_exist], obj=context
    )

    # THEN command should not exit successfully
    assert result.exit_code != EXIT_SUCCESS

    # THEN ERROR log should be printed containing invalid case_id
    assert case_id_does_not_exist in caplog.text
    assert "could not be found" in caplog.text


@pytest.mark.parametrize(
    "workflow",
    [Workflow.RNAFUSION, Workflow.TAXPROFILER, Workflow.TOMTE],
)
def test_metrics_deliver_case(
    cli_runner: CliRunner,
    caplog: LogCaptureFixture,
    workflow: Workflow,
    request: FixtureRequest,
):
    """Test command with a case id and a finished analysis which should execute successfully."""
    caplog.set_level(logging.INFO)
    context: CGConfig = request.getfixturevalue(f"{workflow}_context")

    # GIVEN a case for which we mocked files created after a successful run
    case_id: str = request.getfixturevalue(f"{workflow}_case_id")
    metrics_deliverables_path: Path = request.getfixturevalue(
        f"{workflow}_metrics_deliverables_path"
    )
    request.getfixturevalue(f"{workflow}_mock_deliverable_dir")
    request.getfixturevalue(f"{workflow}_mock_analysis_finish")

    # WHEN invoking the command
    result = cli_runner.invoke(workflow_cli, [workflow, "metrics-deliver", case_id], obj=context)

    # THEN command should execute successfully
    assert result.exit_code == EXIT_SUCCESS

    # THEN metrics deliverable file should be generated
    assert "Writing metrics deliverables file to" in caplog.text
    assert metrics_deliverables_path.is_file()

    # WHEN reading metrics_deliverable_content as yaml
    metrics_deliverables_content = read_yaml(file_path=metrics_deliverables_path)
    assert isinstance(metrics_deliverables_content, dict)

    # THEN the metrics_deliverables file contains the key "metrics"
    assert "metrics" in metrics_deliverables_content.keys()



from cg.models.orders.order import OrderIn
from cg.server.dto.orders.orders_request import OrdersRequest
from cg.server.dto.orders.orders_response import Order as OrderResponse
from cg.server.dto.orders.orders_response import OrdersResponse
from cg.services.orders.order_service.utils import (
    create_order_response,
    create_orders_response,
    order_is_delivered,
)
from cg.services.orders.order_status_service import OrderSummaryService
from cg.services.orders.order_status_service.dto.order_summary import OrderSummary
from cg.store.models import Case, Order
from cg.store.store import Store


class OrderService:
    def __init__(self, store: Store, status_service: OrderSummaryService) -> None:
        self.store = store
        self.summary_service = status_service

    def get_order(self, order_id: int) -> OrderResponse:
        order: Order = self.store.get_order_by_id(order_id)
        summary: OrderSummary = self.summary_service.get_summary(order_id)
        return create_order_response(order=order, summary=summary)

    def get_orders(self, orders_request: OrdersRequest) -> OrdersResponse:
        orders, total_count = self.store.get_orders(orders_request)
        order_ids: list[int] = [order.id for order in orders]
        summaries: list[OrderSummary] = self.summary_service.get_summaries(order_ids)
        return create_orders_response(orders=orders, summaries=summaries, total=total_count)

    def create_order(self, order_data: OrderIn) -> OrderResponse:
        """Creates an order and links it to the given cases."""
        order: Order = self.store.add_order(order_data)
        cases: list[Case] = self.store.get_cases_by_ticket_id(order_data.ticket)
        for case in cases:
            self.store.link_case_to_order(order_id=order.id, case_id=case.id)
        return create_order_response(order)

    def set_delivery(self, order_id: int, delivered: bool) -> OrderResponse:
        order: Order = self.store.update_order_delivery(order_id=order_id, delivered=delivered)
        return create_order_response(order)

    def update_delivered(self, order_id: int, delivered_analyses: int) -> None:
        """Update the delivery status of an order based on the number of delivered analyses."""
        order: Order = self.store.get_order_by_id(order_id)
        case_count: int = len(order.cases)
        if order_is_delivered(case_count=case_count, delivered_analyses=delivered_analyses):
            self.set_delivery(order_id=order_id, delivered=True)
        elif order.is_delivered:
            self.set_delivery(order_id=order_id, delivered=False)



"""Test module cg.cli.upload.nipt"""

import datetime
import logging

from click.testing import CliRunner

from cg.apps.tb.api import TrailblazerAPI
from cg.cli.upload.nipt.base import nipt_upload_all, nipt_upload_case
from cg.constants.constants import Workflow
from cg.meta.upload.nipt import NiptUploadAPI
from cg.models.cg_config import CGConfig
from cg.store.models import Analysis

NIPT_CASE_SUCCESS = "*** NIPT UPLOAD START ***"
NIPT_ALL_SUCCESS = "*** NIPT UPLOAD ALL START ***"
NIPT_STATINA_SUCCESS = "*** Statina UPLOAD START ***"
NIPT_FTP_SUCCESS = "*** NIPT FTP UPLOAD START ***"


class MockStatinaUploadFiles:
    def json(self, *args, **kwargs):
        return ""


def test_nipt_statina_upload_case(
    upload_context: CGConfig, cli_runner: CliRunner, caplog, helpers, mocker
):
    """Tests CLI command to upload a single case"""

    # GIVEN a specified NIPT case that has its analysis stored but is not yet uploaded
    caplog.set_level(logging.DEBUG)

    analysis_obj: Analysis = helpers.add_analysis(store=upload_context.status_db)
    case_id = analysis_obj.case.internal_id
    assert not analysis_obj.upload_started_at
    assert not analysis_obj.uploaded_at

    # WHEN uploading of a specified NIPT case
    mocker.patch.object(NiptUploadAPI, "get_statina_files", return_value=MockStatinaUploadFiles())
    mocker.patch.object(NiptUploadAPI, "upload_to_statina_database")
    mocker.patch.object(NiptUploadAPI, "get_housekeeper_results_file")
    mocker.patch.object(NiptUploadAPI, "get_results_file_path")
    mocker.patch.object(NiptUploadAPI, "upload_to_ftp_server")
    mocker.patch.object(NiptUploadAPI, "flowcell_passed_qc_value", return_value=True)
    mocker.patch.object(TrailblazerAPI, "set_analysis_uploaded")
    result = cli_runner.invoke(
        cli=nipt_upload_case, args=[case_id], obj=upload_context, catch_exceptions=False
    )

    # THEN both the nipt ftp and statina upload should start
    assert NIPT_CASE_SUCCESS in caplog.text
    assert NIPT_STATINA_SUCCESS in caplog.text
    assert NIPT_FTP_SUCCESS in caplog.text

    # THEN set analysis.upload_started_at in the database
    assert analysis_obj.upload_started_at

    # THEN set analysis.uploaded_at in the database
    assert analysis_obj.uploaded_at

    # THEN exit without errors
    assert result.exit_code == 0


def test_nipt_statina_upload_case_dry_run(
    upload_context: CGConfig, cli_runner: CliRunner, caplog, helpers, mocker
):
    """Tests CLI command to upload a single case"""

    # GIVEN a specified NIPT case that has its analysis stored but is not yet uploaded
    caplog.set_level(logging.DEBUG)

    analysis_obj: Analysis = helpers.add_analysis(store=upload_context.status_db)
    case_id = analysis_obj.case.internal_id
    assert not analysis_obj.upload_started_at
    assert not analysis_obj.uploaded_at

    # WHEN uploading a specified NIPT case with dry-run flag set
    mocker.patch.object(NiptUploadAPI, "get_statina_files", return_value=MockStatinaUploadFiles())
    mocker.patch.object(NiptUploadAPI, "upload_to_statina_database")
    mocker.patch.object(NiptUploadAPI, "get_housekeeper_results_file")
    mocker.patch.object(NiptUploadAPI, "get_results_file_path")
    mocker.patch.object(NiptUploadAPI, "upload_to_ftp_server")
    mocker.patch.object(NiptUploadAPI, "flowcell_passed_qc_value", return_value=True)
    result = cli_runner.invoke(
        cli=nipt_upload_case, args=[case_id, "--dry-run"], obj=upload_context
    )

    # THEN both the nipt ftp and statina upload should start
    assert NIPT_CASE_SUCCESS in caplog.text
    assert NIPT_STATINA_SUCCESS in caplog.text
    assert NIPT_FTP_SUCCESS in caplog.text

    # THEN analysis.upload_started_at should not be set in the database
    assert analysis_obj.upload_started_at is None

    # THEN analysis.uploaded_at should not be set in the database
    assert analysis_obj.uploaded_at is None

    # THEN exit without errors
    assert result.exit_code == 0


def test_nipt_statina_upload_auto(
    upload_context: CGConfig, cli_runner: CliRunner, caplog, helpers, mocker
):
    """Tests CLI command to upload a single case"""

    # GIVEN a case ready for upload
    caplog.set_level(logging.DEBUG)

    analysis_obj: Analysis = helpers.add_analysis(
        store=upload_context.status_db,
        completed_at=datetime.datetime.now(),
        workflow=Workflow.FLUFFY,
    )
    assert analysis_obj.completed_at
    assert not analysis_obj.uploaded_at

    # WHEN uploading all NIPT cases
    mocker.patch.object(NiptUploadAPI, "get_statina_files", return_value=MockStatinaUploadFiles())
    mocker.patch.object(NiptUploadAPI, "upload_to_statina_database")
    mocker.patch.object(NiptUploadAPI, "get_housekeeper_results_file")
    mocker.patch.object(NiptUploadAPI, "get_results_file_path")
    mocker.patch.object(NiptUploadAPI, "upload_to_ftp_server")
    mocker.patch.object(NiptUploadAPI, "flowcell_passed_qc_value", return_value=True)
    mocker.patch.object(TrailblazerAPI, "set_analysis_uploaded")

    result = cli_runner.invoke(cli=nipt_upload_all, args=[], obj=upload_context)

    # THEN both the nipt ftp and statina upload should start
    assert NIPT_ALL_SUCCESS in caplog.text
    assert NIPT_STATINA_SUCCESS in caplog.text
    assert NIPT_FTP_SUCCESS in caplog.text

    # THEN set analysis.upload_started_at in the database
    assert analysis_obj.upload_started_at

    # THEN set analysis.uploaded_at in the database
    assert analysis_obj.uploaded_at

    # THEN exit without errors
    assert result.exit_code == 0


def test_nipt_statina_upload_auto_without_analyses(
    upload_context: CGConfig, cli_runner: CliRunner, caplog, helpers, mocker
):
    """Tests CLI command to upload without any analyses to upload."""

    # GIVEN no analyses for upload
    caplog.set_level(logging.DEBUG)

    mocker.patch.object(NiptUploadAPI, "get_all_upload_analyses", return_value=None)

    # WHEN uploading all NIPT cases
    result = cli_runner.invoke(cli=nipt_upload_all, args=[], obj=upload_context)

    # THEN the command should abort without raising an error
    assert result.exit_code == 0
    assert "No analyses found to upload" in caplog.text


def test_nipt_statina_upload_auto_analysis_without_case(
    upload_context: CGConfig, cli_runner: CliRunner, caplog, helpers, mocker
):
    """Tests CLI command to upload a non-existing case."""

    # GIVEN no analyses for upload
    caplog.set_level(logging.DEBUG)

    analysis_obj: Analysis = helpers.add_analysis(
        store=upload_context.status_db,
        completed_at=datetime.datetime.now(),
        workflow=Workflow.FLUFFY,
    )
    analysis_obj.case = None
    mocker.patch.object(NiptUploadAPI, "get_all_upload_analyses", return_value=[analysis_obj])
    # WHEN uploading all NIPT cases
    result = cli_runner.invoke(cli=nipt_upload_all, args=[], obj=upload_context)

    # THEN the command should abort without raising an error
    assert result.exit_code != 0


def test_nipt_statina_upload_auto_dry_run(
    upload_context: CGConfig, cli_runner: CliRunner, caplog, helpers, mocker
):
    """Tests CLI command to upload a single case"""

    # GIVEN a case ready for upload
    caplog.set_level(logging.DEBUG)

    analysis_obj: Analysis = helpers.add_analysis(
        store=upload_context.status_db,
        completed_at=datetime.datetime.now(),
        workflow=Workflow.FLUFFY,
    )
    assert analysis_obj.completed_at
    assert not analysis_obj.uploaded_at

    # WHEN uploading all NIPT cases
    mocker.patch.object(NiptUploadAPI, "get_statina_files", return_value=MockStatinaUploadFiles())
    mocker.patch.object(NiptUploadAPI, "upload_to_statina_database")
    mocker.patch.object(NiptUploadAPI, "get_housekeeper_results_file")
    mocker.patch.object(NiptUploadAPI, "get_results_file_path")
    mocker.patch.object(NiptUploadAPI, "upload_to_ftp_server")
    mocker.patch.object(NiptUploadAPI, "flowcell_passed_qc_value", return_value=True)
    result = cli_runner.invoke(cli=nipt_upload_all, args=["--dry-run"], obj=upload_context)

    # THEN both the nipt ftp and statina upload should start
    assert NIPT_ALL_SUCCESS in caplog.text
    assert NIPT_STATINA_SUCCESS in caplog.text
    assert NIPT_FTP_SUCCESS in caplog.text

    # THEN analysis.upload_started_at should not be set in the database
    assert analysis_obj.upload_started_at is None

    # THEN analysis.uploaded_at should not be set in the database
    assert analysis_obj.uploaded_at is None

    # THEN exit without errors
    assert result.exit_code == 0


def test_nipt_statina_upload_force_failed_case(
    upload_context: CGConfig, cli_runner: CliRunner, caplog, helpers, mocker
):
    """Tests CLI command to upload a single case"""

    # GIVEN a completed NIPT case, not yet uploaded
    caplog.set_level(logging.DEBUG)

    analysis_obj: Analysis = helpers.add_analysis(store=upload_context.status_db)
    case_id = analysis_obj.case.internal_id

    # WHEN uploading of a specified NIPT case AND the qc fails but it forced to upload
    mocker.patch.object(NiptUploadAPI, "get_statina_files", return_value=MockStatinaUploadFiles())
    mocker.patch.object(NiptUploadAPI, "upload_to_statina_database")
    mocker.patch.object(NiptUploadAPI, "get_housekeeper_results_file")
    mocker.patch.object(NiptUploadAPI, "get_results_file_path")
    mocker.patch.object(NiptUploadAPI, "upload_to_ftp_server")
    mocker.patch.object(NiptUploadAPI, "flowcell_passed_qc_value", return_value=False)
    mocker.patch.object(TrailblazerAPI, "set_analysis_uploaded")
    result = cli_runner.invoke(
        cli=nipt_upload_case, args=[case_id, "--force"], obj=upload_context, catch_exceptions=False
    )

    # THEN both the nipt ftp and statina upload should finish successfully
    assert NIPT_CASE_SUCCESS in caplog.text
    assert NIPT_STATINA_SUCCESS in caplog.text
    assert NIPT_FTP_SUCCESS in caplog.text

    # THEN exit without errors
    assert result.exit_code == 0



from enum import StrEnum

from cg.constants.constants import Strandedness
from cg.models.nf_analysis import NextflowSampleSheetEntry


class TomteSampleSheetEntry(NextflowSampleSheetEntry):
    """Tomte sample model is used when building the sample sheet."""

    case_id: str
    strandedness: Strandedness

    @property
    def reformat_sample_content(self) -> list[list[str]]:
        """Reformat sample sheet content as a list of lists, where
        each list represents a line in the final file."""
        return [
            [
                self.case_id,
                self.name,
                fastq_forward_read_path,
                fastq_reverse_read_path,
                str(self.strandedness),
            ]
            for fastq_forward_read_path, fastq_reverse_read_path in zip(
                self.fastq_forward_read_paths, self.fastq_reverse_read_paths
            )
        ]


class TomteSampleSheetHeaders(StrEnum):
    case_id: str = "case"
    name: str = "sample"
    fastq_1: str = "fastq_1"
    fastq_2: str = "fastq_2"
    strandedness: str = "strandedness"

    @classmethod
    def list(cls) -> list[str]:
        return list(map(lambda header: header.value, cls))



"""Test do decompress a spring archive"""

import logging


def test_decompress_spring(populated_decompress_spring_api, compression_files, sample, caplog):
    """Test to compress all fastq files for a sample"""
    caplog.set_level(logging.DEBUG)
    compress_api = populated_decompress_spring_api
    compress_api.crunchy_api.set_dry_run(True)
    # GIVEN a populated compress api
    spring_path = compression_files.spring_file
    # GIVEN that the spring paths exists
    assert spring_path.exists()
    spring_metadata_path = compression_files.spring_metadata_file
    assert spring_metadata_path.exists()
    # GIVEN that the fastq paths does not exist
    compression_files.fastq_first.unlink()
    compression_files.fastq_second.unlink()
    assert not compression_files.fastq_first.exists()
    assert not compression_files.fastq_second.exists()

    # WHEN Decompressing the spring file for the sample
    res = compress_api.decompress_spring(sample)

    # THEN assert compression succeded
    assert res is True
    # THEN assert that the correct information is communicated
    assert f"Decompressing {spring_path} to FASTQ format for sample {sample}" in caplog.text



from pathlib import Path

from cg.io.json import read_json
from cg.meta.workflow.microsalt.metrics_parser.models import QualityMetrics


class MetricsParser:
    @staticmethod
    def parse(file_path: Path) -> QualityMetrics:
        data = read_json(file_path)
        formatted_data = {"samples": data}
        return QualityMetrics.model_validate(formatted_data)



"""Tests for the SampleSheetCreator classes."""

from pathlib import Path

from cg.apps.demultiplex.sample_sheet.read_sample_sheet import get_flow_cell_samples_from_content
from cg.apps.demultiplex.sample_sheet.sample_models import FlowCellSample, FlowCellSampleBCLConvert
from cg.apps.demultiplex.sample_sheet.sample_sheet_creator import SampleSheetCreator


def test_construct_bcl_convert_sheet(
    bcl_convert_sample_sheet_creator: SampleSheetCreator, project_dir: Path
):
    """Test that a created BCL Convert sample sheet has samples."""
    # GIVEN a BCL convert sample sheet creator populated with BCL convert samples
    assert bcl_convert_sample_sheet_creator.lims_samples

    # WHEN building the sample sheet content
    content: list[list[str]] = bcl_convert_sample_sheet_creator.construct_sample_sheet()

    # THEN a correctly formatted sample sheet was created
    samples: list[FlowCellSample] = get_flow_cell_samples_from_content(
        sample_sheet_content=content, sample_type=FlowCellSampleBCLConvert
    )
    assert samples



"""Delivery API bundle fixtures."""

from copy import deepcopy
from pathlib import Path
from typing import Any

import pytest

from cg.constants import SequencingFileTag
from cg.constants.housekeeper_tags import HK_DELIVERY_REPORT_TAG, AlignmentFileTag


@pytest.fixture
def hk_delivery_sample_bundle(
    sample_hk_bundle_no_files: dict[str, Any],
    sample_id: str,
    delivery_fastq_file: Path,
    delivery_spring_file: Path,
) -> dict:
    sample_hk_bundle: dict[str, Any] = deepcopy(sample_hk_bundle_no_files)
    sample_hk_bundle["name"] = sample_id
    sample_hk_bundle["files"] = [
        {
            "archive": False,
            "path": delivery_fastq_file.as_posix(),
            "tags": [SequencingFileTag.FASTQ, sample_id],
        },
        {
            "archive": False,
            "path": delivery_spring_file.as_posix(),
            "tags": [SequencingFileTag.SPRING, sample_id],
        },
    ]
    return sample_hk_bundle


@pytest.fixture
def hk_delivery_another_sample_bundle(
    sample_hk_bundle_no_files: dict[str, Any],
    another_sample_id: str,
    delivery_another_fastq_file: Path,
    delivery_another_spring_file: Path,
) -> dict:
    sample_hk_bundle: dict[str, Any] = deepcopy(sample_hk_bundle_no_files)
    sample_hk_bundle["name"] = another_sample_id
    sample_hk_bundle["files"] = [
        {
            "archive": False,
            "path": delivery_another_fastq_file.as_posix(),
            "tags": [SequencingFileTag.FASTQ, another_sample_id],
        },
        {
            "archive": False,
            "path": delivery_another_spring_file.as_posix(),
            "tags": [SequencingFileTag.SPRING, another_sample_id],
        },
    ]
    return sample_hk_bundle


@pytest.fixture
def hk_delivery_case_bundle(
    case_hk_bundle_no_files: dict[str, Any],
    case_id: str,
    sample_id: str,
    another_sample_id: str,
    delivery_report_file: Path,
    delivery_cram_file: Path,
    delivery_another_cram_file: Path,
) -> dict:
    case_hk_bundle: dict[str, Any] = deepcopy(case_hk_bundle_no_files)
    case_hk_bundle["name"] = case_id
    case_hk_bundle["files"] = [
        {
            "archive": False,
            "path": delivery_report_file.as_posix(),
            "tags": [HK_DELIVERY_REPORT_TAG, case_id],
        },
        {
            "archive": False,
            "path": delivery_cram_file.as_posix(),
            "tags": [AlignmentFileTag.CRAM, sample_id],
        },
        {
            "archive": False,
            "path": delivery_another_cram_file.as_posix(),
            "tags": [AlignmentFileTag.CRAM, another_sample_id],
        },
    ]
    return case_hk_bundle



from enum import Enum
from typing import Callable

from sqlalchemy import func
from sqlalchemy.orm import Query

from cg.store.models import SampleLaneSequencingMetrics


def filter_total_read_count_for_sample(metrics: Query, sample_internal_id: str, **kwargs) -> Query:
    """Return total read count for sample across all lanes."""
    total_reads_query: Query = metrics.with_entities(
        func.sum(SampleLaneSequencingMetrics.sample_total_reads_in_lane)
    ).filter(SampleLaneSequencingMetrics.sample_internal_id == sample_internal_id)
    return total_reads_query


def filter_above_q30_threshold(metrics: Query, q30_threshold: int, **kwargs) -> Query:
    """Filter metrics above Q30 threshold and return the ratio."""
    return metrics.filter(
        SampleLaneSequencingMetrics.sample_base_percentage_passing_q30 > q30_threshold,
    )


def filter_by_flow_cell_sample_internal_id_and_lane(
    metrics: Query, flow_cell_name: str, sample_internal_id: str, lane: int, **kwargs
) -> Query:
    """Filter metrics by flow cell name, sample internal id and lane."""
    return metrics.filter(
        SampleLaneSequencingMetrics.flow_cell_name == flow_cell_name,
        SampleLaneSequencingMetrics.sample_internal_id == sample_internal_id,
        SampleLaneSequencingMetrics.flow_cell_lane_number == lane,
    )


def filter_by_flow_cell_name(metrics: Query, flow_cell_name: str, **kwargs) -> Query:
    """Filter metrics by flow cell name."""
    return metrics.filter(
        SampleLaneSequencingMetrics.flow_cell_name == flow_cell_name,
    )


def filter_by_sample_internal_id(metrics: Query, sample_internal_id: str, **kwargs) -> Query:
    """Filter metrics by sample internal id."""
    return metrics.filter(
        SampleLaneSequencingMetrics.sample_internal_id == sample_internal_id,
    )


class SequencingMetricsFilter(Enum):
    TOTAL_READ_COUNT_FOR_SAMPLE: Callable = filter_total_read_count_for_sample
    BY_FLOW_CELL_SAMPLE_INTERNAL_ID_AND_LANE: Callable = (
        filter_by_flow_cell_sample_internal_id_and_lane
    )
    BY_FLOW_CELL_NAME: Callable = filter_by_flow_cell_name
    BY_SAMPLE_INTERNAL_ID: Callable = filter_by_sample_internal_id
    ABOVE_Q30_THRESHOLD: Callable = filter_above_q30_threshold


def apply_metrics_filter(
    metrics: Query,
    filter_functions: list[Callable],
    sample_internal_id: str | None = None,
    flow_cell_name: str | None = None,
    lane: int | None = None,
    q30_threshold: int | None = None,
) -> Query:
    for filter_function in filter_functions:
        metrics: Query = filter_function(
            metrics=metrics,
            sample_internal_id=sample_internal_id,
            flow_cell_name=flow_cell_name,
            lane=lane,
            q30_threshold=q30_threshold,
        )
    return metrics



from cg.apps.tb.models import TrailblazerAnalysis


class MockTB:
    def __init__(self):
        self.analyses_response = []
        self.get_latest_analysis_response = {}

    def is_latest_analysis_ongoing(self, *args, **kwargs) -> bool:
        return False

    def add_pending_analysis(self, *args, **kwargs) -> None:
        return None

    def add_commit(self, *args, **kwargs) -> None:
        return None

    def has_latest_analysis_started(self, *args, **kwargs) -> bool:
        return False

    def analyses(self, *args, **kwargs) -> list:
        return self.analyses_response

    def get_latest_analysis(self, case_id: str) -> TrailblazerAnalysis | None:
        return self.get_latest_analysis_response.get(case_id)

    def ensure_analyses_response(self, analyses_list: list) -> None:
        self.analyses_response = [
            TrailblazerAnalysis.model_validate(analysis) for analysis in analyses_list
        ]

    def ensure_get_latest_analysis_response(self, analysis_dict: dict) -> None:
        self.get_latest_analysis_response[analysis_dict["family"]] = (
            TrailblazerAnalysis.model_validate(analysis_dict)
        )

    def is_latest_analysis_completed(self, case_id: str):
        return True

    def is_latest_analysis_qc(self, case_id: str):
        return True

    def set_analysis_status(self, case_id: str, status: str):
        return

    def add_comment(self, case_id: str, comment: str):
        return

    def get_summaries(self, order_ids: list[int]):
        return []



"""Tests for coverage meta API"""

from datetime import datetime

from cg.apps.coverage.api import ChanjoAPI
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.meta.upload.coverage import UploadCoverageApi
from cg.store.models import Case
from cg.store.store import Store


class MockAnalysis:
    """Mock Analysis object"""

    def __init__(self, case_obj: Case, started_at: datetime):
        self.case_obj = case_obj
        self.case = case_obj
        self.started_at = started_at

    @property
    def family(self):
        """mock case"""
        return self.case_obj


def test_data(
    coverage_upload_api: UploadCoverageApi,
    analysis_store: Store,
    case_id: str,
    timestamp_yesterday: datetime,
):
    """test getting data for chanjo"""
    # GIVEN a coverage api and an analysis object
    coverage_api = coverage_upload_api
    case_name = case_id
    case_obj = analysis_store.get_case_by_internal_id(internal_id=case_name)
    analysis_obj = MockAnalysis(case_obj=case_obj, started_at=timestamp_yesterday)

    # WHEN using the data method
    results = coverage_api.data(analysis_obj=analysis_obj)

    # THEN this returns the data needed to upload samples to chanjo
    assert results["family"] == case_name
    for sample in results["samples"]:
        assert set(sample.keys()) == set(["coverage", "sample", "sample_name"])


def test_upload(
    chanjo_config: dict,
    populated_housekeeper_api: HousekeeperAPI,
    analysis_store: Store,
    mocker,
    case_id: str,
    timestamp_yesterday: datetime,
):
    """test uploading with chanjo."""
    # GIVEN a coverage api and a data dictionary
    mock_upload = mocker.patch.object(ChanjoAPI, "upload")
    mock_sample = mocker.patch.object(ChanjoAPI, "sample")
    mocker.patch.object(ChanjoAPI, "delete_sample")
    hk_api = populated_housekeeper_api
    chanjo_api = ChanjoAPI(config=chanjo_config)
    coverage_api = UploadCoverageApi(status_api=None, hk_api=hk_api, chanjo_api=chanjo_api)
    family_name = case_id
    case_obj = analysis_store.get_case_by_internal_id(family_name)
    analysis_obj = MockAnalysis(case_obj=case_obj, started_at=timestamp_yesterday)
    data = coverage_api.data(analysis_obj=analysis_obj)

    # WHEN uploading samples in data dictionary
    coverage_api.upload(data=data, replace=True)

    # THEN methods sample, and upload should each have been called three times
    assert mock_upload.call_count == len(data["samples"])
    assert mock_sample.call_count == len(data["samples"])



from cg.meta.orders.sars_cov_2_submitter import SarsCov2Submitter
from cg.models.orders.constants import OrderType
from cg.models.orders.order import OrderIn
from cg.models.orders.sample_base import ControlEnum
from cg.models.orders.samples import SarsCov2Sample
from cg.store.store import Store


def test_order_to_status_control_exists(sarscov2_order_to_submit: dict, base_store: Store):
    # GIVEN sarscov2 order with three samples
    order: OrderIn = OrderIn.parse_obj(sarscov2_order_to_submit, OrderType.SARS_COV_2)

    # WHEN transforming order to status structure
    result: dict = SarsCov2Submitter.order_to_status(order=order)

    # THEN check that control is in the result
    sample: dict
    for sample in result.get("samples"):
        assert "control" in sample


def test_order_to_status_control_has_input_value(sarscov2_order_to_submit: dict, base_store: Store):
    # GIVEN sarscov2 order with three samples with control value set
    control_value = ControlEnum.positive
    order: OrderIn = OrderIn.parse_obj(sarscov2_order_to_submit, OrderType.SARS_COV_2)
    sample: SarsCov2Sample
    for sample in order.samples:
        sample.control: ControlEnum = control_value

    # WHEN transforming order to status structure
    result: dict = SarsCov2Submitter.order_to_status(order=order)

    # THEN check that control is in the result
    sample: dict
    for sample in result.get("samples"):
        assert control_value in sample.get("control")



"""This module contains constants for the BCLConvert metrics files."""

from enum import StrEnum

from cg.constants.demultiplexing import (
    BclConverter,
    SampleSheetBcl2FastqSections,
    SampleSheetBCLConvertSections,
)


class BclConvertQualityMetricsColumnNames(StrEnum):
    """Column names for the BCLConvert quality metrics file."""

    LANE: str = "Lane"
    SAMPLE_INTERNAL_ID: str = "SampleID"
    MEAN_QUALITY_SCORE_Q30: str = "Mean Quality Score (PF)"
    Q30_BASES_PERCENT: str = "% Q30"


SAMPLE_SHEET_HEADER: dict[str, str] = {
    BclConverter.BCL2FASTQ: ",".join(
        [
            column
            for column in SampleSheetBcl2FastqSections.Data.column_names()
            if column != SampleSheetBcl2FastqSections.Data.INDEX_2.value
        ]
    ),
    BclConverter.BCLCONVERT: ",".join(
        [
            column
            for column in SampleSheetBCLConvertSections.Data.column_names()
            if column != SampleSheetBcl2FastqSections.Data.INDEX_2.value
        ]
    ),
}


class BclConvertDemuxMetricsColumnNames(StrEnum):
    """Column names for the BCLConvert demultiplexing metrics file."""

    LANE: str = "Lane"
    SAMPLE_INTERNAL_ID: str = "SampleID"
    READ_PAIR_COUNT: str = "# Reads"


DEMUX_METRICS_FILE_NAME: str = "Demultiplex_Stats.csv"
QUALITY_METRICS_FILE_NAME: str = "Quality_Metrics.csv"
ADAPTER_METRICS_FILE_NAME: str = "Adapter_Metrics.csv"
